@@ -1,5 +1,7 @@
+// SPDX-License-Identifier: Apache-2.0
 /*
  * umoci: Umoci Modifies Open Containers' Images
+ * Copyright (C) 2016-2025 SUSE LLC
  * Copyright (C) 2018 Cisco Systems
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
@@ -15,12 +17,17 @@
  * limitations under the License.
  */
 
+// Package umoci provides the top-level Go APIs for umoci functionality. Note
+// that the umoci Go API is not yet considered stable -- we will aim to
+// document any breaking changes in our changelogs but users should verify that
+// updates will not break their usage of umoci.
 package umoci
 
 import (
+	"fmt"
+
 	"github.com/opencontainers/umoci/oci/cas/dir"
 	"github.com/opencontainers/umoci/oci/casext"
-	"github.com/pkg/errors"
 )
 
 // OpenLayout opens an existing OCI image layout, and fails if it does not
@@ -29,7 +36,7 @@ func OpenLayout(imagePath string) (casext.Engine, error) {
 	// Get a reference to the CAS.
 	engine, err := dir.Open(imagePath)
 	if err != nil {
-		return casext.Engine{}, errors.Wrap(err, "open CAS")
+		return casext.Engine{}, fmt.Errorf("open CAS: %w", err)
 	}
 
 	return casext.NewEngine(engine), nil
