@@ -1,6 +1,7 @@
+// SPDX-License-Identifier: Apache-2.0
 /*
  * umoci: Umoci Modifies Open Containers' Images
- * Copyright (C) 2016-2020 SUSE LLC
+ * Copyright (C) 2016-2025 SUSE LLC
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -19,21 +20,25 @@ package main
 
 import (
 	"context"
+	"errors"
+	"fmt"
 	"strings"
 	"time"
 
 	"github.com/apex/log"
 	ispec "github.com/opencontainers/image-spec/specs-go/v1"
+	"github.com/urfave/cli"
+
 	"github.com/opencontainers/umoci/mutate"
 	"github.com/opencontainers/umoci/oci/cas/dir"
 	"github.com/opencontainers/umoci/oci/casext"
 	igen "github.com/opencontainers/umoci/oci/config/generate"
-	"github.com/pkg/errors"
-	"github.com/urfave/cli"
+	"github.com/opencontainers/umoci/pkg/funchelpers"
 )
 
 // FIXME: We should also implement a raw mode that just does modifications of
-//        JSON blobs (allowing this all to be used outside of our build setup).
+//
+//	JSON blobs (allowing this all to be used outside of our build setup).
 var configCommand = uxHistory(uxTag(cli.Command{
 	Name:  "config",
 	Usage: "modifies the image configuration of an OCI image",
@@ -51,13 +56,13 @@ image.`,
 	// Verify the metadata.
 	Before: func(ctx *cli.Context) error {
 		if ctx.NArg() != 0 {
-			return errors.Errorf("invalid number of positional arguments: expected none")
+			return errors.New("invalid number of positional arguments: expected none")
 		}
 		if _, ok := ctx.App.Metadata["--image-path"]; !ok {
-			return errors.Errorf("missing mandatory argument: --image")
+			return errors.New("missing mandatory argument: --image")
 		}
 		if _, ok := ctx.App.Metadata["--image-tag"]; !ok {
-			return errors.Errorf("missing mandatory argument: --image")
+			return errors.New("missing mandatory argument: --image")
 		}
 		return nil
 	},
@@ -125,69 +130,69 @@ func fromImage(image ispec.Image) (ispec.ImageConfig, mutate.Meta) {
 func parseKV(input string) (string, string, error) {
 	parts := strings.SplitN(input, "=", 2)
 	if len(parts) != 2 {
-		return "", "", errors.Errorf("must contain '=': %s", input)
+		return "", "", fmt.Errorf("must contain '=': %s", input)
 	}
 
 	name, value := parts[0], parts[1]
 	if name == "" {
-		return "", "", errors.Errorf("must have non-empty name: %s", input)
+		return "", "", fmt.Errorf("must have non-empty name: %s", input)
 	}
 	return name, value, nil
 }
 
-func config(ctx *cli.Context) error {
-	imagePath := ctx.App.Metadata["--image-path"].(string)
-	fromName := ctx.App.Metadata["--image-tag"].(string)
+func config(ctx *cli.Context) (Err error) {
+	imagePath := mustFetchMeta[string](ctx, "--image-path")
+	fromName := mustFetchMeta[string](ctx, "--image-tag")
 
 	// By default we clobber the old tag.
 	tagName := fromName
-	if val, ok := ctx.App.Metadata["--tag"]; ok {
-		tagName = val.(string)
+	if val, ok := fetchMeta[string](ctx, "--tag"); ok {
+		tagName = val
 	}
 
 	// Get a reference to the CAS.
 	engine, err := dir.Open(imagePath)
 	if err != nil {
-		return errors.Wrap(err, "open CAS")
+		return fmt.Errorf("open CAS: %w", err)
 	}
 	engineExt := casext.NewEngine(engine)
-	defer engine.Close()
+	defer funchelpers.VerifyClose(&Err, engine)
 
 	fromDescriptorPaths, err := engineExt.ResolveReference(context.Background(), fromName)
 	if err != nil {
-		return errors.Wrap(err, "get descriptor")
+		return fmt.Errorf("get descriptor: %w", err)
 	}
 	if len(fromDescriptorPaths) == 0 {
-		return errors.Errorf("tag not found: %s", fromName)
+		return fmt.Errorf("tag not found: %s", fromName)
 	}
 	if len(fromDescriptorPaths) != 1 {
 		// TODO: Handle this more nicely.
-		return errors.Errorf("tag is ambiguous: %s", fromName)
+		return fmt.Errorf("tag is ambiguous: %s", fromName)
 	}
 
 	mutator, err := mutate.New(engine, fromDescriptorPaths[0])
 	if err != nil {
-		return errors.Wrap(err, "create mutator for manifest")
+		return fmt.Errorf("create mutator for manifest: %w", err)
 	}
 
-	imageConfig, err := mutator.Config(context.Background())
+	config, err := mutator.Config(context.Background())
 	if err != nil {
-		return errors.Wrap(err, "get base config")
+		return fmt.Errorf("get base config: %w", err)
 	}
 
 	imageMeta, err := mutator.Meta(context.Background())
 	if err != nil {
-		return errors.Wrap(err, "get base metadata")
+		return fmt.Errorf("get base metadata: %w", err)
 	}
 
 	annotations, err := mutator.Annotations(context.Background())
 	if err != nil {
-		return errors.Wrap(err, "get base annotations")
+		return fmt.Errorf("get base annotations: %w", err)
 	}
 
-	g, err := igen.NewFromImage(toImage(imageConfig, imageMeta))
+	g, err := igen.NewFromImage(toImage(config.Config, imageMeta))
 	if err != nil {
-		return errors.Wrap(err, "create new generator")
+		return fmt.Errorf("create new generator: %w", err)
 	}
 
 	if ctx.IsSet("clear") {
@@ -204,14 +209,14 @@ func config(ctx *cli.Context) error {
 			case "config.volume":
 				g.ClearConfigVolumes()
 			case "rootfs.diffids":
-				//g.ClearRootfsDiffIDs()
-				return errors.Errorf("--clear=rootfs.diffids is not safe")
+				// g.ClearRootfsDiffIDs()
+				return errors.New("--clear=rootfs.diffids is not safe")
 			case "config.cmd":
 				g.ClearConfigCmd()
 			case "config.entrypoint":
 				g.ClearConfigEntrypoint()
 			default:
-				return errors.Errorf("unknown key to --clear: %s", key)
+				return fmt.Errorf("unknown key to --clear: %s", key)
 			}
 		}
 	}
@@ -220,7 +225,7 @@ func config(ctx *cli.Context) error {
 		// How do we handle other formats?
 		created, err := time.Parse(igen.ISO8601, ctx.String("created"))
 		if err != nil {
-			return errors.Wrap(err, "parse --created")
+			return fmt.Errorf("parse --created: %w", err)
 		}
 		g.SetCreated(created)
 	}
@@ -251,7 +256,7 @@ func config(ctx *cli.Context) error {
 		for _, env := range ctx.StringSlice("config.env") {
 			name, value, err := parseKV(env)
 			if err != nil {
-				return errors.Wrap(err, "config.env")
+				return fmt.Errorf("config.env: %w", err)
 			}
 			g.AddConfigEnv(name, value)
 		}
@@ -273,7 +278,7 @@ func config(ctx *cli.Context) error {
 		for _, label := range ctx.StringSlice("config.label") {
 			name, value, err := parseKV(label)
 			if err != nil {
-				return errors.Wrap(err, "config.label")
+				return fmt.Errorf("config.label: %w", err)
 			}
 			g.AddConfigLabel(name, value)
 		}
@@ -308,7 +313,7 @@ func config(ctx *cli.Context) error {
 		if ctx.IsSet("history.created") {
 			created, err := time.Parse(igen.ISO8601, ctx.String("history.created"))
 			if err != nil {
-				return errors.Wrap(err, "parsing --history.created")
+				return fmt.Errorf("parsing --history.created: %w", err)
 			}
 			history.Created = &created
 		}
@@ -319,18 +324,18 @@ func config(ctx *cli.Context) error {
 
 	newConfig, newMeta := fromImage(g.Image())
 	if err := mutator.Set(context.Background(), newConfig, newMeta, annotations, history); err != nil {
-		return errors.Wrap(err, "set modified configuration")
+		return fmt.Errorf("set modified configuration: %w", err)
 	}
 
 	newDescriptorPath, err := mutator.Commit(context.Background())
 	if err != nil {
-		return errors.Wrap(err, "commit mutated image")
+		return fmt.Errorf("commit mutated image: %w", err)
 	}
 
 	log.Infof("new image manifest created: %s->%s", newDescriptorPath.Root().Digest, newDescriptorPath.Descriptor().Digest)
 
 	if err := engineExt.UpdateReference(context.Background(), tagName, newDescriptorPath.Root()); err != nil {
-		return errors.Wrap(err, "add new tag")
+		return fmt.Errorf("add new tag: %w", err)
 	}
 
 	log.Infof("created new tag for image manifest: %s", tagName)
