@@ -1,6 +1,7 @@
+// SPDX-License-Identifier: Apache-2.0
 /*
  * umoci: Umoci Modifies Open Containers' Images
- * Copyright (C) 2016-2020 SUSE LLC
+ * Copyright (C) 2016-2025 SUSE LLC
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -18,13 +19,14 @@
 package main
 
 import (
+	"errors"
 	"fmt"
 	"os"
 
 	"github.com/apex/log"
-	"github.com/opencontainers/umoci/oci/cas/dir"
-	"github.com/pkg/errors"
 	"github.com/urfave/cli"
+
+	"github.com/opencontainers/umoci/oci/cas/dir"
 )
 
 var initCommand = cli.Command{
@@ -43,7 +45,7 @@ commands.`,
 
 	Before: func(ctx *cli.Context) error {
 		if ctx.NArg() != 0 {
-			return errors.Errorf("invalid number of positional arguments: expected none")
+			return errors.New("invalid number of positional arguments: expected none")
 		}
 		return nil
 	},
@@ -52,17 +54,17 @@ commands.`,
 }
 
 func initLayout(ctx *cli.Context) error {
-	imagePath := ctx.App.Metadata["--image-path"].(string)
+	imagePath := mustFetchMeta[string](ctx, "--image-path")
 
-	if _, err := os.Stat(imagePath); !os.IsNotExist(err) {
+	if _, err := os.Stat(imagePath); !errors.Is(err, os.ErrNotExist) {
 		if err == nil {
 			err = fmt.Errorf("path already exists: %s", imagePath)
 		}
-		return errors.Wrap(err, "image layout creation")
+		return fmt.Errorf("image layout creation: %w", err)
 	}
 
 	if err := dir.Create(imagePath); err != nil {
-		return errors.Wrap(err, "image layout creation")
+		return fmt.Errorf("image layout creation: %w", err)
 	}
 
 	log.Infof("created new OCI image: %s", imagePath)
