@@ -1,6 +1,7 @@
+// SPDX-License-Identifier: Apache-2.0
 /*
  * umoci: Umoci Modifies Open Containers' Images
- * Copyright (C) 2016-2020 SUSE LLC
+ * Copyright (C) 2016-2025 SUSE LLC
  * Copyright (C) 2018 Cisco Systems
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
@@ -20,21 +21,24 @@ package main
 
 import (
 	"context"
+	"fmt"
 	"time"
 
 	"github.com/apex/log"
 	ispec "github.com/opencontainers/image-spec/specs-go/v1"
+	"github.com/urfave/cli"
+
 	"github.com/opencontainers/umoci"
 	"github.com/opencontainers/umoci/mutate"
 	"github.com/opencontainers/umoci/oci/cas/dir"
 	"github.com/opencontainers/umoci/oci/casext"
+	"github.com/opencontainers/umoci/oci/casext/blobcompress"
 	igen "github.com/opencontainers/umoci/oci/config/generate"
 	"github.com/opencontainers/umoci/oci/layer"
-	"github.com/pkg/errors"
-	"github.com/urfave/cli"
+	"github.com/opencontainers/umoci/pkg/funchelpers"
 )
 
-var insertCommand = uxRemap(uxHistory(uxTag(cli.Command{
+var insertCommand = uxCompress(uxRemap(uxHistory(uxTag(cli.Command{
 	Name:  "insert",
 	Usage: "insert content into an OCI image",
 	ArgsUsage: `--image <image-path>[:<tag>] [--opaque] <source> <target>
@@ -87,11 +91,11 @@ Some examples:
 			numArgs = 1
 		}
 		if ctx.NArg() != numArgs {
-			return errors.Errorf("invalid number of positional arguments: expected %d", numArgs)
+			return fmt.Errorf("invalid number of positional arguments: expected %d", numArgs)
 		}
 		for idx, args := range ctx.Args() {
 			if args == "" {
-				return errors.Errorf("invalid positional argument %d: arguments cannot be empty", idx)
+				return fmt.Errorf("invalid positional argument %d: arguments cannot be empty", idx)
 			}
 		}
 
@@ -107,44 +111,49 @@ Some examples:
 		ctx.App.Metadata["--target-path"] = targetPath
 		return nil
 	},
-})))
+}))))
 
-func insert(ctx *cli.Context) error {
-	imagePath := ctx.App.Metadata["--image-path"].(string)
-	fromName := ctx.App.Metadata["--image-tag"].(string)
-	sourcePath := ctx.App.Metadata["--source-path"].(string)
-	targetPath := ctx.App.Metadata["--target-path"].(string)
+func insert(ctx *cli.Context) (Err error) {
+	imagePath := mustFetchMeta[string](ctx, "--image-path")
+	fromName := mustFetchMeta[string](ctx, "--image-tag")
+	sourcePath := mustFetchMeta[string](ctx, "--source-path")
+	targetPath := mustFetchMeta[string](ctx, "--target-path")
+
+	var compressAlgo blobcompress.Algorithm
+	if algo, ok := fetchMeta[blobcompress.Algorithm](ctx, "--compress"); ok {
+		compressAlgo = algo
+	}
 
 	// By default we clobber the old tag.
 	tagName := fromName
-	if val, ok := ctx.App.Metadata["--tag"]; ok {
-		tagName = val.(string)
+	if val, ok := fetchMeta[string](ctx, "--tag"); ok {
+		tagName = val
 	}
 
 	// Get a reference to the CAS.
 	engine, err := dir.Open(imagePath)
 	if err != nil {
-		return errors.Wrap(err, "open CAS")
+		return fmt.Errorf("open CAS: %w", err)
 	}
 	engineExt := casext.NewEngine(engine)
-	defer engine.Close()
+	defer funchelpers.VerifyClose(&Err, engine)
 
 	descriptorPaths, err := engineExt.ResolveReference(context.Background(), fromName)
 	if err != nil {
-		return errors.Wrap(err, "get descriptor")
+		return fmt.Errorf("get descriptor: %w", err)
 	}
 	if len(descriptorPaths) == 0 {
-		return errors.Errorf("tag not found: %s", fromName)
+		return fmt.Errorf("tag not found: %s", fromName)
 	}
 	if len(descriptorPaths) != 1 {
 		// TODO: Handle this more nicely.
-		return errors.Errorf("tag is ambiguous: %s", fromName)
+		return fmt.Errorf("tag is ambiguous: %s", fromName)
 	}
 
 	// Create the mutator.
 	mutator, err := mutate.New(engine, descriptorPaths[0])
 	if err != nil {
-		return errors.Wrap(err, "create mutator for base image")
+		return fmt.Errorf("create mutator for base image: %w", err)
 	}
 
 	var meta umoci.Meta
@@ -156,9 +165,13 @@ func insert(ctx *cli.Context) error {
 		return err
 	}
 
-	packOptions := layer.RepackOptions{MapOptions: meta.MapOptions}
+	packOptions := layer.RepackOptions{
+		OnDiskFormat: layer.DirRootfs{
+			MapOptions: meta.MapOptions,
+		},
+	}
 	reader := layer.GenerateInsertLayer(sourcePath, targetPath, ctx.IsSet("opaque"), &packOptions)
-	defer reader.Close()
+	defer funchelpers.VerifyClose(&Err, reader)
 
 	var history *ispec.History
 	if !ctx.Bool("no-history") {
@@ -179,7 +192,7 @@ func insert(ctx *cli.Context) error {
 		if ctx.IsSet("history.created") {
 			created, err := time.Parse(igen.ISO8601, ctx.String("history.created"))
 			if err != nil {
-				return errors.Wrap(err, "parsing --history.created")
+				return fmt.Errorf("parsing --history.created: %w", err)
 			}
 			history.Created = &created
 		}
@@ -188,21 +201,19 @@ func insert(ctx *cli.Context) error {
 		}
 	}
 
-	// TODO: We should add a flag to allow for a new layer to be made
-	//       non-distributable.
-	if _, err := mutator.Add(context.Background(), ispec.MediaTypeImageLayer, reader, history, mutate.GzipCompressor); err != nil {
-		return errors.Wrap(err, "add diff layer")
+	if _, err := mutator.Add(context.Background(), ispec.MediaTypeImageLayer, reader, history, compressAlgo, nil); err != nil {
+		return fmt.Errorf("add diff layer: %w", err)
 	}
 
 	newDescriptorPath, err := mutator.Commit(context.Background())
 	if err != nil {
-		return errors.Wrap(err, "commit mutated image")
+		return fmt.Errorf("commit mutated image: %w", err)
 	}
 
 	log.Infof("new image manifest created: %s->%s", newDescriptorPath.Root().Digest, newDescriptorPath.Descriptor().Digest)
 
 	if err := engineExt.UpdateReference(context.Background(), tagName, newDescriptorPath.Root()); err != nil {
-		return errors.Wrap(err, "add new tag")
+		return fmt.Errorf("add new tag: %w", err)
 	}
 	log.Infof("updated tag for image manifest: %s", tagName)
 	return nil
