@@ -1,6 +1,7 @@
+// SPDX-License-Identifier: Apache-2.0
 /*
  * umoci: Umoci Modifies Open Containers' Images
- * Copyright (C) 2016-2020 SUSE LLC
+ * Copyright (C) 2016-2025 SUSE LLC
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -19,21 +20,25 @@ package main
 
 import (
 	"context"
+	"errors"
+	"fmt"
 	"os"
 	"time"
 
 	"github.com/apex/log"
 	ispec "github.com/opencontainers/image-spec/specs-go/v1"
+	"github.com/urfave/cli"
+
 	"github.com/opencontainers/umoci"
 	"github.com/opencontainers/umoci/mutate"
 	"github.com/opencontainers/umoci/oci/cas/dir"
 	"github.com/opencontainers/umoci/oci/casext"
+	"github.com/opencontainers/umoci/oci/casext/blobcompress"
 	igen "github.com/opencontainers/umoci/oci/config/generate"
-	"github.com/pkg/errors"
-	"github.com/urfave/cli"
+	"github.com/opencontainers/umoci/pkg/funchelpers"
 )
 
-var rawAddLayerCommand = uxHistory(uxTag(cli.Command{
+var rawAddLayerCommand = uxCompress(uxHistory(uxTag(cli.Command{
 	Name:  "add-layer",
 	Usage: "add a layer archive verbatim to an image",
 	ArgsUsage: `--image <image-path>[:<tag>] <new-layer.tar>
@@ -56,25 +61,30 @@ only supports uncompressed archives.`,
 
 	Before: func(ctx *cli.Context) error {
 		if ctx.NArg() != 1 {
-			return errors.Errorf("invalid number of positional arguments: expected <newlayer.tar>")
+			return errors.New("invalid number of positional arguments: expected <newlayer.tar>")
 		}
 		if ctx.Args().First() == "" {
-			return errors.Errorf("<new-layer.tar> path cannot be empty")
+			return errors.New("<new-layer.tar> path cannot be empty")
 		}
 		ctx.App.Metadata["newlayer"] = ctx.Args().First()
 		return nil
 	},
-}))
+})))
 
-func rawAddLayer(ctx *cli.Context) error {
-	imagePath := ctx.App.Metadata["--image-path"].(string)
-	fromName := ctx.App.Metadata["--image-tag"].(string)
-	newLayerPath := ctx.App.Metadata["newlayer"].(string)
+func rawAddLayer(ctx *cli.Context) (Err error) {
+	imagePath := mustFetchMeta[string](ctx, "--image-path")
+	fromName := mustFetchMeta[string](ctx, "--image-tag")
+	newLayerPath := mustFetchMeta[string](ctx, "newlayer")
+
+	var compressAlgo blobcompress.Algorithm
+	if algo, ok := fetchMeta[blobcompress.Algorithm](ctx, "--compress"); ok {
+		compressAlgo = algo
+	}
 
 	// Overide the from tag by default, otherwise use the one specified.
 	tagName := fromName
-	if overrideTagName, ok := ctx.App.Metadata["--tag"]; ok {
-		tagName = overrideTagName.(string)
+	if overrideTagName, ok := fetchMeta[string](ctx, "--tag"); ok {
+		tagName = overrideTagName
 	}
 
 	var meta umoci.Meta
@@ -83,45 +93,45 @@ func rawAddLayer(ctx *cli.Context) error {
 	// Get a reference to the CAS.
 	engine, err := dir.Open(imagePath)
 	if err != nil {
-		return errors.Wrap(err, "open CAS")
+		return fmt.Errorf("open CAS: %w", err)
 	}
 	engineExt := casext.NewEngine(engine)
-	defer engine.Close()
+	defer funchelpers.VerifyClose(&Err, engine)
 
 	fromDescriptorPaths, err := engineExt.ResolveReference(context.Background(), fromName)
 	if err != nil {
-		return errors.Wrap(err, "get descriptor")
+		return fmt.Errorf("get descriptor: %w", err)
 	}
 	if len(fromDescriptorPaths) == 0 {
-		return errors.Errorf("tag not found: %s", fromName)
+		return fmt.Errorf("tag not found: %s", fromName)
 	}
 	if len(fromDescriptorPaths) != 1 {
 		// TODO: Handle this more nicely.
-		return errors.Errorf("tag is ambiguous: %s", fromName)
+		return fmt.Errorf("tag is ambiguous: %s", fromName)
 	}
 	meta.From = fromDescriptorPaths[0]
 
 	// Create the mutator.
 	mutator, err := mutate.New(engine, meta.From)
 	if err != nil {
-		return errors.Wrap(err, "create mutator for base image")
+		return fmt.Errorf("create mutator for base image: %w", err)
 	}
 
 	newLayer, err := os.Open(newLayerPath)
 	if err != nil {
-		return errors.Wrap(err, "open new layer archive")
+		return fmt.Errorf("open new layer archive: %w", err)
 	}
 	if fi, err := newLayer.Stat(); err != nil {
-		return errors.Wrap(err, "stat new layer archive")
+		return fmt.Errorf("stat new layer archive: %w", err)
 	} else if fi.IsDir() {
-		return errors.Errorf("new layer archive is a directory")
+		return errors.New("new layer archive is a directory")
 	}
 	// TODO: Verify that the layer is actually uncompressed.
-	defer newLayer.Close()
+	defer funchelpers.VerifyClose(&Err, newLayer)
 
 	imageMeta, err := mutator.Meta(context.Background())
 	if err != nil {
-		return errors.Wrap(err, "get image metadata")
+		return fmt.Errorf("get image metadata: %w", err)
 	}
 
 	var history *ispec.History
@@ -144,7 +154,7 @@ func rawAddLayer(ctx *cli.Context) error {
 		if ctx.IsSet("history.created") {
 			created, err := time.Parse(igen.ISO8601, ctx.String("history.created"))
 			if err != nil {
-				return errors.Wrap(err, "parsing --history.created")
+				return fmt.Errorf("parsing --history.created: %w", err)
 			}
 			history.Created = &created
 		}
@@ -153,21 +163,19 @@ func rawAddLayer(ctx *cli.Context) error {
 		}
 	}
 
-	// TODO: We should add a flag to allow for a new layer to be made
-	//       non-distributable.
-	if _, err := mutator.Add(context.Background(), ispec.MediaTypeImageLayer, newLayer, history, mutate.GzipCompressor); err != nil {
-		return errors.Wrap(err, "add diff layer")
+	if _, err := mutator.Add(context.Background(), ispec.MediaTypeImageLayer, newLayer, history, compressAlgo, nil); err != nil {
+		return fmt.Errorf("add diff layer: %w", err)
 	}
 
 	newDescriptorPath, err := mutator.Commit(context.Background())
 	if err != nil {
-		return errors.Wrap(err, "commit mutated image")
+		return fmt.Errorf("commit mutated image: %w", err)
 	}
 
 	log.Infof("new image manifest created: %s->%s", newDescriptorPath.Root().Digest, newDescriptorPath.Descriptor().Digest)
 
 	if err := engineExt.UpdateReference(context.Background(), tagName, newDescriptorPath.Root()); err != nil {
-		return errors.Wrap(err, "add new tag")
+		return fmt.Errorf("add new tag: %w", err)
 	}
 
 	log.Infof("created new tag for image manifest: %s", tagName)
