@@ -1,6 +1,7 @@
+// SPDX-License-Identifier: Apache-2.0
 /*
  * umoci: Umoci Modifies Open Containers' Images
- * Copyright (C) 2016-2020 SUSE LLC
+ * Copyright (C) 2016-2025 SUSE LLC
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -19,16 +20,18 @@ package main
 
 import (
 	"context"
+	"errors"
 	"fmt"
 
 	"github.com/apex/log"
 	ispec "github.com/opencontainers/image-spec/specs-go/v1"
+	"github.com/urfave/cli"
+
 	"github.com/opencontainers/umoci"
 	"github.com/opencontainers/umoci/oci/cas/dir"
 	"github.com/opencontainers/umoci/oci/casext"
 	"github.com/opencontainers/umoci/oci/layer"
-	"github.com/pkg/errors"
-	"github.com/urfave/cli"
+	"github.com/opencontainers/umoci/pkg/funchelpers"
 )
 
 var rawUnpackCommand = uxRemap(cli.Command{
@@ -54,22 +57,21 @@ is the destination to unpack the image to.`,
 
 	Before: func(ctx *cli.Context) error {
 		if ctx.NArg() != 1 {
-			return errors.Errorf("invalid number of positional arguments: expected <rootfs>")
+			return errors.New("invalid number of positional arguments: expected <rootfs>")
 		}
 		if ctx.Args().First() == "" {
-			return errors.Errorf("rootfs path cannot be empty")
+			return errors.New("rootfs path cannot be empty")
 		}
 		ctx.App.Metadata["rootfs"] = ctx.Args().First()
 		return nil
 	},
 })
 
-func rawUnpack(ctx *cli.Context) error {
-	imagePath := ctx.App.Metadata["--image-path"].(string)
-	fromName := ctx.App.Metadata["--image-tag"].(string)
-	rootfsPath := ctx.App.Metadata["rootfs"].(string)
+func rawUnpack(ctx *cli.Context) (Err error) {
+	imagePath := mustFetchMeta[string](ctx, "--image-path")
+	fromName := mustFetchMeta[string](ctx, "--image-tag")
+	rootfsPath := mustFetchMeta[string](ctx, "rootfs")
 
-	var unpackOptions layer.UnpackOptions
 	var meta umoci.Meta
 	meta.Version = umoci.MetaVersion
 
@@ -80,38 +82,42 @@ func rawUnpack(ctx *cli.Context) error {
 		return err
 	}
 
-	unpackOptions.KeepDirlinks = ctx.Bool("keep-dirlinks")
-	unpackOptions.MapOptions = meta.MapOptions
+	unpackOptions := layer.UnpackOptions{
+		OnDiskFormat: layer.DirRootfs{
+			MapOptions: meta.MapOptions,
+		},
+		KeepDirlinks: ctx.Bool("keep-dirlinks"),
+	}
 
 	// Get a reference to the CAS.
 	engine, err := dir.Open(imagePath)
 	if err != nil {
-		return errors.Wrap(err, "open CAS")
+		return fmt.Errorf("open CAS: %w", err)
 	}
 	engineExt := casext.NewEngine(engine)
-	defer engine.Close()
+	defer funchelpers.VerifyClose(&Err, engine)
 
 	fromDescriptorPaths, err := engineExt.ResolveReference(context.Background(), fromName)
 	if err != nil {
-		return errors.Wrap(err, "get descriptor")
+		return fmt.Errorf("get descriptor: %w", err)
 	}
 	if len(fromDescriptorPaths) == 0 {
-		return errors.Errorf("tag is not found: %s", fromName)
+		return fmt.Errorf("tag is not found: %s", fromName)
 	}
 	if len(fromDescriptorPaths) != 1 {
 		// TODO: Handle this more nicely.
-		return errors.Errorf("tag is ambiguous: %s", fromName)
+		return fmt.Errorf("tag is ambiguous: %s", fromName)
 	}
 	meta.From = fromDescriptorPaths[0]
 
 	manifestBlob, err := engineExt.FromDescriptor(context.Background(), meta.From.Descriptor())
 	if err != nil {
-		return errors.Wrap(err, "get manifest")
+		return fmt.Errorf("get manifest: %w", err)
 	}
-	defer manifestBlob.Close()
+	defer funchelpers.VerifyClose(&Err, manifestBlob)
 
 	if manifestBlob.Descriptor.MediaType != ispec.MediaTypeImageManifest {
-		return errors.Wrap(fmt.Errorf("descriptor does not point to ispec.MediaTypeImageManifest: not implemented: %s", manifestBlob.Descriptor.MediaType), "invalid --image tag")
+		return fmt.Errorf("invalid --image tag: descriptor does not point to ispec.MediaTypeImageManifest: not implemented: %s", manifestBlob.Descriptor.MediaType)
 	}
 
 	log.WithFields(log.Fields{
@@ -124,12 +130,12 @@ func rawUnpack(ctx *cli.Context) error {
 	manifest, ok := manifestBlob.Data.(ispec.Manifest)
 	if !ok {
 		// Should _never_ be reached.
-		return errors.Errorf("[internal error] unknown manifest blob type: %s", manifestBlob.Descriptor.MediaType)
+		return fmt.Errorf("[internal error] unknown manifest blob type: %s", manifestBlob.Descriptor.MediaType)
 	}
 
 	log.Warnf("unpacking rootfs ...")
 	if err := layer.UnpackRootfs(context.Background(), engineExt, rootfsPath, manifest, &unpackOptions); err != nil {
-		return errors.Wrap(err, "create rootfs")
+		return fmt.Errorf("create rootfs: %w", err)
 	}
 	log.Warnf("... done")
 
