@@ -1,6 +1,7 @@
+// SPDX-License-Identifier: Apache-2.0
 /*
  * umoci: Umoci Modifies Open Containers' Images
- * Copyright (C) 2016-2020 SUSE LLC
+ * Copyright (C) 2016-2025 SUSE LLC
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -19,13 +20,15 @@ package main
 
 import (
 	"context"
+	"errors"
 	"fmt"
 
 	"github.com/apex/log"
+	"github.com/urfave/cli"
+
 	"github.com/opencontainers/umoci/oci/cas/dir"
 	"github.com/opencontainers/umoci/oci/casext"
-	"github.com/pkg/errors"
-	"github.com/urfave/cli"
+	"github.com/opencontainers/umoci/pkg/funchelpers"
 )
 
 var tagAddCommand = cli.Command{
@@ -43,49 +46,49 @@ the tag and "<new-tag>" is the new name of the tag.`,
 
 	Before: func(ctx *cli.Context) error {
 		if ctx.NArg() != 1 {
-			return errors.Errorf("invalid number of positional arguments: expected <new-tag>")
+			return errors.New("invalid number of positional arguments: expected <new-tag>")
 		}
 		if ctx.Args().First() == "" {
-			return errors.Errorf("new tag cannot be empty")
+			return errors.New("new tag cannot be empty")
 		}
 		if !casext.IsValidReferenceName(ctx.Args().First()) {
-			return errors.Errorf("new tag is an invalid reference")
+			return errors.New("new tag is an invalid reference")
 		}
 		ctx.App.Metadata["new-tag"] = ctx.Args().First()
 		return nil
 	},
 }
 
-func tagAdd(ctx *cli.Context) error {
-	imagePath := ctx.App.Metadata["--image-path"].(string)
-	fromName := ctx.App.Metadata["--image-tag"].(string)
-	tagName := ctx.App.Metadata["new-tag"].(string)
+func tagAdd(ctx *cli.Context) (Err error) {
+	imagePath := mustFetchMeta[string](ctx, "--image-path")
+	fromName := mustFetchMeta[string](ctx, "--image-tag")
+	tagName := mustFetchMeta[string](ctx, "new-tag")
 
 	// Get a reference to the CAS.
 	engine, err := dir.Open(imagePath)
 	if err != nil {
-		return errors.Wrap(err, "open CAS")
+		return fmt.Errorf("open CAS: %w", err)
 	}
 	engineExt := casext.NewEngine(engine)
-	defer engine.Close()
+	defer funchelpers.VerifyClose(&Err, engine)
 
 	// Get original descriptor.
 	descriptorPaths, err := engineExt.ResolveReference(context.Background(), fromName)
 	if err != nil {
-		return errors.Wrap(err, "get descriptor")
+		return fmt.Errorf("get descriptor: %w", err)
 	}
 	if len(descriptorPaths) == 0 {
-		return errors.Errorf("tag not found: %s", fromName)
+		return fmt.Errorf("tag not found: %s", fromName)
 	}
 	if len(descriptorPaths) != 1 {
 		// TODO: Handle this more nicely.
-		return errors.Errorf("tag is ambiguous: %s", fromName)
+		return fmt.Errorf("tag is ambiguous: %s", fromName)
 	}
 	descriptor := descriptorPaths[0].Descriptor()
 
 	// Add it.
 	if err := engineExt.UpdateReference(context.Background(), tagName, descriptor); err != nil {
-		return errors.Wrap(err, "put reference")
+		return fmt.Errorf("put reference: %w", err)
 	}
 
 	log.Infof("created new tag: %q -> %q", tagName, fromName)
@@ -107,7 +110,7 @@ tag to remove.`,
 
 	Before: func(ctx *cli.Context) error {
 		if ctx.NArg() != 0 {
-			return errors.Errorf("invalid number of positional arguments: expected none")
+			return errors.New("invalid number of positional arguments: expected none")
 		}
 		return nil
 	},
@@ -115,21 +118,21 @@ tag to remove.`,
 	Action: tagRemove,
 }
 
-func tagRemove(ctx *cli.Context) error {
-	imagePath := ctx.App.Metadata["--image-path"].(string)
-	tagName := ctx.App.Metadata["--image-tag"].(string)
+func tagRemove(ctx *cli.Context) (Err error) {
+	imagePath := mustFetchMeta[string](ctx, "--image-path")
+	tagName := mustFetchMeta[string](ctx, "--image-tag")
 
 	// Get a reference to the CAS.
 	engine, err := dir.Open(imagePath)
 	if err != nil {
-		return errors.Wrap(err, "open CAS")
+		return fmt.Errorf("open CAS: %w", err)
 	}
 	engineExt := casext.NewEngine(engine)
-	defer engine.Close()
+	defer funchelpers.VerifyClose(&Err, engine)
 
 	// Remove it.
 	if err := engineExt.DeleteReference(context.Background(), tagName); err != nil {
-		return errors.Wrap(err, "delete reference")
+		return fmt.Errorf("delete reference: %w", err)
 	}
 
 	log.Infof("removed tag: %s", tagName)
@@ -152,7 +155,7 @@ line. See umoci-stat(1) to get more information about each tagged image.`,
 
 	Before: func(ctx *cli.Context) error {
 		if ctx.NArg() != 0 {
-			return errors.Errorf("invalid number of positional arguments: expected none")
+			return errors.New("invalid number of positional arguments: expected none")
 		}
 		return nil
 	},
@@ -160,20 +163,20 @@ line. See umoci-stat(1) to get more information about each tagged image.`,
 	Action: tagList,
 }
 
-func tagList(ctx *cli.Context) error {
-	imagePath := ctx.App.Metadata["--image-path"].(string)
+func tagList(ctx *cli.Context) (Err error) {
+	imagePath := mustFetchMeta[string](ctx, "--image-path")
 
 	// Get a reference to the CAS.
 	engine, err := dir.Open(imagePath)
 	if err != nil {
-		return errors.Wrap(err, "open CAS")
+		return fmt.Errorf("open CAS: %w", err)
 	}
 	engineExt := casext.NewEngine(engine)
-	defer engine.Close()
+	defer funchelpers.VerifyClose(&Err, engine)
 
 	names, err := engineExt.ListReferences(context.Background())
 	if err != nil {
-		return errors.Wrap(err, "list references")
+		return fmt.Errorf("list references: %w", err)
 	}
 
 	for _, name := range names {
