@@ -1,6 +1,7 @@
+// SPDX-License-Identifier: Apache-2.0
 /*
  * umoci: Umoci Modifies Open Containers' Images
- * Copyright (C) 2016-2020 SUSE LLC
+ * Copyright (C) 2016-2025 SUSE LLC
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -15,13 +16,16 @@
  * limitations under the License.
  */
 
+// Package dir implements the basic local directory-backed layout backend for
+// the OCI content-addressible store.
 package dir
 
 import (
 	"context"
 	"encoding/json"
+	"errors"
+	"fmt"
 	"io"
-	"io/ioutil"
 	"os"
 	"path/filepath"
 
@@ -29,10 +33,12 @@ import (
 	"github.com/opencontainers/go-digest"
 	imeta "github.com/opencontainers/image-spec/specs-go"
 	ispec "github.com/opencontainers/image-spec/specs-go/v1"
-	"github.com/opencontainers/umoci/oci/cas"
-	"github.com/opencontainers/umoci/pkg/hardening"
-	"github.com/pkg/errors"
 	"golang.org/x/sys/unix"
+
+	"github.com/opencontainers/umoci/oci/cas"
+	"github.com/opencontainers/umoci/pkg/funchelpers"
+	"github.com/opencontainers/umoci/pkg/hardening"
+	"github.com/opencontainers/umoci/pkg/system"
 )
 
 const (
@@ -58,14 +64,14 @@ const (
 // of the OCI image. The digest must be of the form algorithm:hex.
 func blobPath(digest digest.Digest) (string, error) {
 	if err := digest.Validate(); err != nil {
-		return "", errors.Wrapf(err, "invalid digest: %q", digest)
+		return "", fmt.Errorf("invalid digest: %q: %w", digest, err)
 	}
 
 	algo := digest.Algorithm()
 	hash := digest.Hex()
 
 	if algo != cas.BlobAlgorithm {
-		return "", errors.Errorf("unsupported algorithm: %q", algo)
+		return "", fmt.Errorf("unsupported algorithm: %q", algo)
 	}
 
 	return filepath.Join(blobDirectory, algo.String(), hash), nil
@@ -79,9 +85,9 @@ type dirEngine struct {
 
 func (e *dirEngine) ensureTempDir() error {
 	if e.temp == "" {
-		tempDir, err := ioutil.TempDir(e.path, ".umoci-")
+		tempDir, err := os.MkdirTemp(e.path, ".umoci-")
 		if err != nil {
-			return errors.Wrap(err, "create tempdir")
+			return fmt.Errorf("create tempdir: %w", err)
 		}
 
 		// We get an advisory lock to ensure that GC() won't delete our
@@ -90,10 +96,10 @@ func (e *dirEngine) ensureTempDir() error {
 
 		e.tempFile, err = os.Open(tempDir)
 		if err != nil {
-			return errors.Wrap(err, "open tempdir for lock")
+			return fmt.Errorf("open tempdir for lock: %w", err)
 		}
 		if err := unix.Flock(int(e.tempFile.Fd()), unix.LOCK_EX|unix.LOCK_NB); err != nil {
-			return errors.Wrap(err, "lock tempdir")
+			return fmt.Errorf("lock tempdir: %w", err)
 		}
 
 		e.temp = tempDir
@@ -103,23 +109,23 @@ func (e *dirEngine) ensureTempDir() error {
 
 // verify ensures that the image is valid.
 func (e *dirEngine) validate() error {
-	content, err := ioutil.ReadFile(filepath.Join(e.path, layoutFile))
+	content, err := os.ReadFile(filepath.Join(e.path, layoutFile))
 	if err != nil {
-		if os.IsNotExist(err) {
+		if errors.Is(err, os.ErrNotExist) {
 			err = cas.ErrInvalid
 		}
-		return errors.Wrap(err, "read oci-layout")
+		return fmt.Errorf("read oci-layout: %w", err)
 	}
 
 	var ociLayout ispec.ImageLayout
 	if err := json.Unmarshal(content, &ociLayout); err != nil {
-		return errors.Wrap(err, "parse oci-layout")
+		return fmt.Errorf("parse oci-layout: %w", err)
 	}
 
 	// XXX: Currently the meaning of this field is not adequately defined by
 	//      the spec, nor is the "official" value determined by the spec.
 	if ociLayout.Version != ImageLayoutVersion {
-		return errors.Wrap(cas.ErrInvalid, "layout version is not supported")
+		return fmt.Errorf("layout version is not supported: %w", cas.ErrInvalid)
 	}
 
 	// Check that "blobs" and "index.json" exist in the image.
@@ -127,21 +133,21 @@ func (e *dirEngine) validate() error {
 	//        directory (with no subdirectories) and that refs *only* contains
 	//        files (optionally also making sure they're all JSON descriptors).
 	if fi, err := os.Stat(filepath.Join(e.path, blobDirectory)); err != nil {
-		if os.IsNotExist(err) {
+		if errors.Is(err, os.ErrNotExist) {
 			err = cas.ErrInvalid
 		}
-		return errors.Wrap(err, "check blobdir")
+		return fmt.Errorf("check blobdir: %w", err)
 	} else if !fi.IsDir() {
-		return errors.Wrap(cas.ErrInvalid, "blobdir is not a directory")
+		return fmt.Errorf("blobdir is not a directory: %w", cas.ErrInvalid)
 	}
 
 	if fi, err := os.Stat(filepath.Join(e.path, indexFile)); err != nil {
-		if os.IsNotExist(err) {
+		if errors.Is(err, os.ErrNotExist) {
 			err = cas.ErrInvalid
 		}
-		return errors.Wrap(err, "check index")
+		return fmt.Errorf("check index: %w", err)
 	} else if fi.IsDir() {
-		return errors.Wrap(cas.ErrInvalid, "index is a directory")
+		return fmt.Errorf("index is a directory: %w", cas.ErrInvalid)
 	}
 
 	return nil
@@ -150,91 +156,126 @@ func (e *dirEngine) validate() error {
 // PutBlob adds a new blob to the image. This is idempotent; a nil error
 // means that "the content is stored at DIGEST" without implying "because
 // of this PutBlob() call".
-func (e *dirEngine) PutBlob(ctx context.Context, reader io.Reader) (digest.Digest, int64, error) {
+func (e *dirEngine) PutBlob(_ context.Context, reader io.Reader) (_ digest.Digest, _ int64, Err error) {
 	if err := e.ensureTempDir(); err != nil {
-		return "", -1, errors.Wrap(err, "ensure tempdir")
+		return "", -1, fmt.Errorf("ensure tempdir: %w", err)
 	}
 
 	digester := cas.BlobAlgorithm.Digester()
 
 	// We copy this into a temporary file because we need to get the blob hash,
 	// but also to avoid half-writing an invalid blob.
-	fh, err := ioutil.TempFile(e.temp, "blob-")
+	fh, err := os.CreateTemp(e.temp, "blob-")
 	if err != nil {
-		return "", -1, errors.Wrap(err, "create temporary blob")
+		return "", -1, fmt.Errorf("create temporary blob: %w", err)
 	}
 	tempPath := fh.Name()
-	defer fh.Close()
+	defer funchelpers.VerifyClose(&Err, fh)
 
 	writer := io.MultiWriter(fh, digester.Hash())
-	size, err := io.Copy(writer, reader)
+	size, err := system.Copy(writer, reader)
 	if err != nil {
-		return "", -1, errors.Wrap(err, "copy to temporary blob")
+		return "", -1, fmt.Errorf("copy to temporary blob: %w", err)
 	}
-	if err := fh.Close(); err != nil {
-		return "", -1, errors.Wrap(err, "close temporary blob")
+	if err := fh.Sync(); err != nil {
+		return "", -1, fmt.Errorf("sync temporary blob: %w", err)
 	}
 
 	// Get the digest.
 	path, err := blobPath(digester.Digest())
 	if err != nil {
-		return "", -1, errors.Wrap(err, "compute blob name")
+		return "", -1, fmt.Errorf("compute blob name: %w", err)
 	}
 
 	// Move the blob to its correct path.
 	path = filepath.Join(e.path, path)
 	if err := os.Rename(tempPath, path); err != nil {
-		return "", -1, errors.Wrap(err, "rename temporary blob")
+		return "", -1, fmt.Errorf("rename temporary blob: %w", err)
 	}
 
-	return digester.Digest(), int64(size), nil
+	return digester.Digest(), size, nil
 }
 
 // GetBlob returns a reader for retrieving a blob from the image, which the
-// caller must Close(). Returns os.ErrNotExist if the digest is not found.
-func (e *dirEngine) GetBlob(ctx context.Context, digest digest.Digest) (io.ReadCloser, error) {
+// caller must Close(). Returns ErrNotExist if the digest is not found.
+//
+// This function will return a VerifiedReadCloser, meaning that you must call
+// Close() and check the error returned from Close() in order to ensure that
+// the hash of the blob is verified.
+//
+// Please note that calling Close() on the returned blob will read the entire
+// from disk and hash it (even if you didn't read any bytes before calling
+// Close), so if you wish to only check if a blob exists you should use
+// StatBlob() instead.
+func (e *dirEngine) GetBlob(_ context.Context, digest digest.Digest) (io.ReadCloser, error) {
 	path, err := blobPath(digest)
 	if err != nil {
-		return nil, errors.Wrap(err, "compute blob path")
+		return nil, fmt.Errorf("compute blob path: %w", err)
 	}
 	fh, err := os.Open(filepath.Join(e.path, path))
+	if err != nil {
+		return nil, fmt.Errorf("open blob: %w", err)
+	}
 	return &hardening.VerifiedReadCloser{
 		Reader:         fh,
 		ExpectedDigest: digest,
 		ExpectedSize:   int64(-1), // We don't know the expected size.
-	}, errors.Wrap(err, "open blob")
+	}, nil
+}
+
+// StatBlob returns whether the specified blob exists in the image. Returns
+// false if the blob doesn't exist, true if it does, or an error if any error
+// occurred.
+//
+// NOTE: In future this API may change to return additional information.
+func (e *dirEngine) StatBlob(_ context.Context, digest digest.Digest) (bool, error) {
+	path, err := blobPath(digest)
+	if err != nil {
+		return false, fmt.Errorf("compute blob path: %w", err)
+	}
+	_, err = os.Stat(path)
+	if errors.Is(err, os.ErrNotExist) {
+		return false, nil
+	}
+	if err != nil {
+		return false, fmt.Errorf("stat blob path: %w", err)
+	}
+	return true, nil
 }
 
 // PutIndex sets the index of the OCI image to the given index, replacing the
 // previously existing index. This operation is atomic; any readers attempting
 // to access the OCI image while it is being modified will only ever see the
 // new or old index.
-func (e *dirEngine) PutIndex(ctx context.Context, index ispec.Index) error {
+func (e *dirEngine) PutIndex(_ context.Context, index ispec.Index) (Err error) {
 	if err := e.ensureTempDir(); err != nil {
-		return errors.Wrap(err, "ensure tempdir")
+		return fmt.Errorf("ensure tempdir: %w", err)
 	}
 
+	// Make sure the index has the mediatype field set.
+	index.MediaType = ispec.MediaTypeImageIndex
+
 	// We copy this into a temporary index to ensure the atomicity of this
 	// operation.
-	fh, err := ioutil.TempFile(e.temp, "index-")
+	fh, err := os.CreateTemp(e.temp, "index-")
 	if err != nil {
-		return errors.Wrap(err, "create temporary index")
+		return fmt.Errorf("create temporary index: %w", err)
 	}
 	tempPath := fh.Name()
-	defer fh.Close()
+	defer funchelpers.VerifyClose(&Err, fh)
 
 	// Encode the index.
 	if err := json.NewEncoder(fh).Encode(index); err != nil {
-		return errors.Wrap(err, "write temporary index")
+		return fmt.Errorf("write temporary index: %w", err)
 	}
-	if err := fh.Close(); err != nil {
-		return errors.Wrap(err, "close temporary index")
+	if err := fh.Sync(); err != nil {
+		return fmt.Errorf("sync temporary index: %w", err)
 	}
 
 	// Move the blob to its correct path.
 	path := filepath.Join(e.path, indexFile)
 	if err := os.Rename(tempPath, path); err != nil {
-		return errors.Wrap(err, "rename temporary index")
+		return fmt.Errorf("rename temporary index: %w", err)
 	}
 	return nil
 }
@@ -248,18 +289,18 @@ func (e *dirEngine) PutIndex(ctx context.Context, index ispec.Index) error {
 // handling nested indexes. casext.Engine provides a wrapper for cas.Engine
 // that implements various reference resolution functions that should work for
 // most users.
-func (e *dirEngine) GetIndex(ctx context.Context) (ispec.Index, error) {
-	content, err := ioutil.ReadFile(filepath.Join(e.path, indexFile))
+func (e *dirEngine) GetIndex(_ context.Context) (ispec.Index, error) {
+	content, err := os.ReadFile(filepath.Join(e.path, indexFile))
 	if err != nil {
-		if os.IsNotExist(err) {
+		if errors.Is(err, os.ErrNotExist) {
 			err = cas.ErrInvalid
 		}
-		return ispec.Index{}, errors.Wrap(err, "read index")
+		return ispec.Index{}, fmt.Errorf("read index: %w", err)
 	}
 
 	var index ispec.Index
 	if err := json.Unmarshal(content, &index); err != nil {
-		return ispec.Index{}, errors.Wrap(err, "parse index")
+		return ispec.Index{}, fmt.Errorf("parse index: %w", err)
 	}
 
 	return index, nil
@@ -268,21 +309,21 @@ func (e *dirEngine) GetIndex(ctx context.Context) (ispec.Index, error) {
 // DeleteBlob removes a blob from the image. This is idempotent; a nil
 // error means "the content is not in the store" without implying "because
 // of this DeleteBlob() call".
-func (e *dirEngine) DeleteBlob(ctx context.Context, digest digest.Digest) error {
+func (e *dirEngine) DeleteBlob(_ context.Context, digest digest.Digest) error {
 	path, err := blobPath(digest)
 	if err != nil {
-		return errors.Wrap(err, "compute blob path")
+		return fmt.Errorf("compute blob path: %w", err)
 	}
 
 	err = os.Remove(filepath.Join(e.path, path))
-	if err != nil && !os.IsNotExist(err) {
-		return errors.Wrap(err, "remove blob")
+	if err != nil && !errors.Is(err, os.ErrNotExist) {
+		return fmt.Errorf("remove blob: %w", err)
 	}
 	return nil
 }
 
 // ListBlobs returns the set of blob digests stored in the image.
-func (e *dirEngine) ListBlobs(ctx context.Context) ([]digest.Digest, error) {
+func (e *dirEngine) ListBlobs(_ context.Context) ([]digest.Digest, error) {
 	digests := []digest.Digest{}
 	blobDir := filepath.Join(e.path, blobDirectory, cas.BlobAlgorithm.String())
 
@@ -291,13 +332,12 @@ func (e *dirEngine) ListBlobs(ctx context.Context) ([]digest.Digest, error) {
 		if path == blobDir {
 			return nil
 		}
-
 		// XXX: Do we need to handle multiple-directory-deep cases?
 		digest := digest.NewDigestFromHex(cas.BlobAlgorithm.String(), filepath.Base(path))
 		digests = append(digests, digest)
 		return nil
 	}); err != nil {
-		return nil, errors.Wrap(err, "walk blobdir")
+		return nil, fmt.Errorf("walk blobdir: %w", err)
 	}
 
 	return digests, nil
@@ -310,11 +350,11 @@ func (e *dirEngine) Clean(ctx context.Context) error {
 	// Remove every .umoci directory that isn't flocked.
 	matches, err := filepath.Glob(filepath.Join(e.path, ".umoci-*"))
 	if err != nil {
-		return errors.Wrap(err, "glob .umoci-*")
+		return fmt.Errorf("glob .umoci-*: %w", err)
 	}
 	for _, path := range matches {
 		err = e.cleanPath(ctx, path)
-		if err != nil && err != filepath.SkipDir {
+		if err != nil && !errors.Is(err, filepath.SkipDir) {
 			return err
 		}
 	}
@@ -322,21 +362,23 @@ func (e *dirEngine) Clean(ctx context.Context) error {
 	return nil
 }
 
-func (e *dirEngine) cleanPath(ctx context.Context, path string) error {
+func (e *dirEngine) cleanPath(_ context.Context, path string) (Err error) {
 	cfh, err := os.Open(path)
-	if err != nil && !os.IsNotExist(err) {
-		return errors.Wrap(err, "open for locking")
+	if err != nil && !errors.Is(err, os.ErrNotExist) {
+		return fmt.Errorf("open for locking: %w", err)
 	}
-	defer cfh.Close()
+	defer funchelpers.VerifyClose(&Err, cfh)
 
 	if err := unix.Flock(int(cfh.Fd()), unix.LOCK_EX|unix.LOCK_NB); err != nil {
 		// If we fail to get a flock(2) then it's probably already locked,
 		// so we shouldn't touch it.
 		return filepath.SkipDir
 	}
-	defer unix.Flock(int(cfh.Fd()), unix.LOCK_UN)
+	defer funchelpers.VerifyError(&Err, func() error {
+		return unix.Flock(int(cfh.Fd()), unix.LOCK_UN)
+	})
 
-	if err := os.RemoveAll(path); os.IsNotExist(err) {
+	if err := os.RemoveAll(path); errors.Is(err, os.ErrNotExist) {
 		return nil // somebody else beat us to it
 	} else if err != nil {
 		log.Warnf("failed to remove %s: %v", path, err)
@@ -352,13 +394,13 @@ func (e *dirEngine) cleanPath(ctx context.Context, path string) error {
 func (e *dirEngine) Close() error {
 	if e.temp != "" {
 		if err := unix.Flock(int(e.tempFile.Fd()), unix.LOCK_UN); err != nil {
-			return errors.Wrap(err, "unlock tempdir")
+			return fmt.Errorf("unlock tempdir: %w", err)
 		}
 		if err := e.tempFile.Close(); err != nil {
-			return errors.Wrap(err, "close tempdir")
+			return fmt.Errorf("close tempdir: %w", err)
 		}
 		if err := os.RemoveAll(e.temp); err != nil {
-			return errors.Wrap(err, "remove tempdir")
+			return fmt.Errorf("remove tempdir: %w", err)
 		}
 	}
 	return nil
@@ -373,7 +415,7 @@ func Open(path string) (cas.Engine, error) {
 	}
 
 	if err := engine.validate(); err != nil {
-		return nil, errors.Wrap(err, "validate")
+		return nil, fmt.Errorf("validate: %w", err)
 	}
 
 	return engine, nil
@@ -382,53 +424,54 @@ func Open(path string) (cas.Engine, error) {
 // Create creates a new OCI image layout at the given path. If the path already
 // exists, os.ErrExist is returned. However, all of the parent components of
 // the path will be created if necessary.
-func Create(path string) error {
+func Create(path string) (Err error) {
 	// We need to fail if path already exists, but we first create all of the
 	// parent paths.
 	dir := filepath.Dir(path)
 	if dir != "." {
-		if err := os.MkdirAll(dir, 0755); err != nil {
-			return errors.Wrap(err, "mkdir parent")
+		if err := os.MkdirAll(dir, 0o755); err != nil {
+			return fmt.Errorf("mkdir parent: %w", err)
 		}
 	}
-	if err := os.Mkdir(path, 0755); err != nil {
-		return errors.Wrap(err, "mkdir")
+	if err := os.Mkdir(path, 0o755); err != nil {
+		return fmt.Errorf("mkdir: %w", err)
 	}
 
 	// Create the necessary directories and "oci-layout" file.
-	if err := os.Mkdir(filepath.Join(path, blobDirectory), 0755); err != nil {
-		return errors.Wrap(err, "mkdir blobdir")
+	if err := os.Mkdir(filepath.Join(path, blobDirectory), 0o755); err != nil {
+		return fmt.Errorf("mkdir blobdir: %w", err)
 	}
-	if err := os.Mkdir(filepath.Join(path, blobDirectory, cas.BlobAlgorithm.String()), 0755); err != nil {
-		return errors.Wrap(err, "mkdir algorithm")
+	if err := os.Mkdir(filepath.Join(path, blobDirectory, cas.BlobAlgorithm.String()), 0o755); err != nil {
+		return fmt.Errorf("mkdir algorithm: %w", err)
 	}
 
 	indexFh, err := os.Create(filepath.Join(path, indexFile))
 	if err != nil {
-		return errors.Wrap(err, "create index.json")
+		return fmt.Errorf("create index.json: %w", err)
 	}
-	defer indexFh.Close()
+	defer funchelpers.VerifyClose(&Err, indexFh)
 
 	defaultIndex := ispec.Index{
 		Versioned: imeta.Versioned{
 			SchemaVersion: 2, // FIXME: This is hardcoded at the moment.
 		},
+		MediaType: ispec.MediaTypeImageIndex,
 	}
 	if err := json.NewEncoder(indexFh).Encode(defaultIndex); err != nil {
-		return errors.Wrap(err, "encode index.json")
+		return fmt.Errorf("encode index.json: %w", err)
 	}
 
 	layoutFh, err := os.Create(filepath.Join(path, layoutFile))
 	if err != nil {
-		return errors.Wrap(err, "create oci-layout")
+		return fmt.Errorf("create oci-layout: %w", err)
 	}
-	defer layoutFh.Close()
+	defer funchelpers.VerifyClose(&Err, layoutFh)
 
 	ociLayout := ispec.ImageLayout{
 		Version: ImageLayoutVersion,
 	}
 	if err := json.NewEncoder(layoutFh).Encode(ociLayout); err != nil {
-		return errors.Wrap(err, "encode oci-layout")
+		return fmt.Errorf("encode oci-layout: %w", err)
 	}
 
 	// Everything is now set up.
