@@ -1,6 +1,7 @@
+// SPDX-License-Identifier: Apache-2.0
 /*
  * umoci: Umoci Modifies Open Containers' Images
- * Copyright (C) 2016-2020 SUSE LLC
+ * Copyright (C) 2016-2025 SUSE LLC
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -24,7 +25,10 @@ import (
 	"github.com/apex/log"
 	"github.com/opencontainers/go-digest"
 	ispec "github.com/opencontainers/image-spec/specs-go/v1"
+
 	"github.com/opencontainers/umoci/oci/cas"
+	"github.com/opencontainers/umoci/oci/casext/mediatype"
+	"github.com/opencontainers/umoci/pkg/funchelpers"
 )
 
 // childDescriptors is a wrapper around MapDescriptors which just creates a
@@ -87,7 +91,7 @@ func (d DescriptorPath) Descriptor() ispec.Descriptor {
 
 // ErrSkipDescriptor is a special error returned by WalkFunc which will cause
 // Walk to not recurse into the descriptor currently being evaluated by
-// WalkFunc.  This interface is roughly equivalent to filepath.SkipDir.
+// WalkFunc. This interface is roughly equivalent to filepath.SkipDir.
 var ErrSkipDescriptor = errors.New("[internal] do not recurse into descriptor")
 
 // WalkFunc is the type of function passed to Walk. It will be a called on each
@@ -98,7 +102,8 @@ var ErrSkipDescriptor = errors.New("[internal] do not recurse into descriptor")
 // caller.
 //
 // TODO: Also provide Blob to WalkFunc so that callers don't need to load blobs
-//       more than once. This is quite important for remote CAS implementations.
+//
+//	more than once. This is quite important for remote CAS implementations.
 type WalkFunc func(descriptorPath DescriptorPath) error
 
 func (ws *walkState) recurse(ctx context.Context, descriptorPath DescriptorPath) (Err error) {
@@ -111,7 +116,7 @@ func (ws *walkState) recurse(ctx context.Context, descriptorPath DescriptorPath)
 
 	// Run walkFunc.
 	if err := ws.walkFunc(descriptorPath); err != nil {
-		if err == ErrSkipDescriptor {
+		if errors.Is(err, ErrSkipDescriptor) {
 			return nil
 		}
 		return err
@@ -119,24 +124,33 @@ func (ws *walkState) recurse(ctx context.Context, descriptorPath DescriptorPath)
 
 	// Get blob to recurse into.
 	descriptor := descriptorPath.Descriptor()
+
+	// Since FromDescriptor gives us a full VerifiedReadCloser (meaning that
+	// Close is expensive if we don't read any bytes), we should only try to
+	// recurse into this thing if we actually can parse it.
+	if mediatype.GetParser(descriptor.MediaType) == nil {
+		log.Infof("skipping walk into non-parseable media-type %v of blob %v", descriptor.MediaType, descriptor.Digest)
+		return nil
+	}
+
+	// Recurse into the blob now.
 	blob, err := ws.engine.FromDescriptor(ctx, descriptor)
 	if err != nil {
 		// Ignore cases where the descriptor points to an object we don't know
 		// how to parse.
-		if err == cas.ErrUnknownType {
+		if errors.Is(err, cas.ErrUnknownType) {
 			log.Infof("skipping walk into unknown media-type %v of blob %v", descriptor.MediaType, descriptor.Digest)
 			return nil
 		}
 		return err
 	}
-	defer func() {
-		if err := blob.Close(); err != nil {
+	defer funchelpers.VerifyError(&Err, func() error {
+		err := blob.Close()
+		if err != nil {
 			log.Warnf("during recursion blob %v had error on Close: %v", descriptor.Digest, err)
-			if Err == nil {
-				Err = err
-			}
 		}
-	}()
+		return err
+	})
 
 	// Recurse into children.
 	for _, child := range childDescriptors(blob.Data) {
@@ -185,7 +199,7 @@ func (e Engine) reachable(ctx context.Context, root ispec.Descriptor) ([]digest.
 	}); err != nil {
 		return nil, err
 	}
-	var reachable []digest.Digest
+	reachable := make([]digest.Digest, 0, len(seen))
 	for node := range seen {
 		reachable = append(reachable, node)
 	}
