@@ -1,6 +1,7 @@
+// SPDX-License-Identifier: Apache-2.0
 /*
  * umoci: Umoci Modifies Open Containers' Images
- * Copyright (C) 2016-2020 SUSE LLC
+ * Copyright (C) 2016-2025 SUSE LLC
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -19,66 +20,31 @@ package layer
 
 import (
 	"archive/tar"
+	"errors"
+	"fmt"
 	"io"
 	"os"
 	"path/filepath"
 	"strings"
 
 	"github.com/apex/log"
+	"golang.org/x/sys/unix"
+
 	"github.com/opencontainers/umoci/pkg/fseval"
-	"github.com/opencontainers/umoci/pkg/testutils"
-	"github.com/pkg/errors"
+	"github.com/opencontainers/umoci/pkg/funchelpers"
+	"github.com/opencontainers/umoci/pkg/system"
 )
 
-// ignoreXattrs is a list of xattr names that should be ignored when
-// creating a new image layer, because they are host-specific and/or would be a
-// bad idea to unpack. They are also excluded from Lclearxattr when extracting
-// an archive.
-// XXX: Maybe we should make this configurable so users can manually blacklist
-//      (or even whitelist) xattrs that they actually want included? Like how
-//      GNU tar's xattr setup works.
-var ignoreXattrs = map[string]struct{}{
-	// SELinux doesn't allow you to set SELinux policies generically. They're
-	// also host-specific. So just ignore them during extraction.
-	"security.selinux": {},
-
-	// NFSv4 ACLs are very system-specific and shouldn't be touched by us, nor
-	// should they be included in images.
-	"system.nfs4_acl": {},
-
-	// In order to support overlayfs whiteout mode, we shouldn't un-set
-	// this after we've set it when writing out the whiteouts.
-	"trusted.overlay.opaque": {},
-
-	// We don't want to these xattrs into the image, because they're only
-	// relevant based on how the build overlay is constructed and will not
-	// be true on the target system once the image is unpacked (e.g. inodes
-	// might be different, impure status won't be true, etc.).
-	"trusted.overlay.redirect": {},
-	"trusted.overlay.origin":   {},
-	"trusted.overlay.impure":   {},
-	"trusted.overlay.nlink":    {},
-	"trusted.overlay.upper":    {},
-	"trusted.overlay.metacopy": {},
-}
-
-func init() {
-	// For test purposes we add a fake forbidden attribute that an unprivileged
-	// user can easily write to (and thus we can test it).
-	if testutils.IsTestBinary() {
-		ignoreXattrs["user.UMOCI:forbidden_xattr"] = struct{}{}
-	}
-}
-
 // tarGenerator is a helper for generating layer diff tars. It should be noted
 // that when using tarGenerator.Add{Path,Whiteout} it is recommended to do it
 // in lexicographic order.
 type tarGenerator struct {
 	tw *tar.Writer
 
-	// mapOptions is the set of mapping options for modifying entries before
-	// they're added to the layer.
-	mapOptions MapOptions
+	// onDiskFmt is the on-disk format that was used for the already-extracted
+	// directory we are operating on. [OverlayfsRootfs] will cause
+	// overlayfs-style whiteouts to be converted to OCI-style whiteouts.
+	onDiskFmt OnDiskFormat
 
 	// Hardlink mapping.
 	inodes map[uint64]string
@@ -92,17 +58,19 @@ type tarGenerator struct {
 
 // newTarGenerator creates a new tarGenerator using the provided writer as the
 // output writer.
-func newTarGenerator(w io.Writer, opt MapOptions) *tarGenerator {
+func newTarGenerator(w io.Writer, opt *RepackOptions) *tarGenerator {
+	opt = opt.fill()
+
 	fsEval := fseval.Default
-	if opt.Rootless {
+	if opt.MapOptions().Rootless {
 		fsEval = fseval.Rootless
 	}
 
 	return &tarGenerator{
-		tw:         tar.NewWriter(w),
-		mapOptions: opt,
-		inodes:     map[uint64]string{},
-		fsEval:     fsEval,
+		tw:        tar.NewWriter(w),
+		onDiskFmt: opt.OnDiskFormat,
+		inodes:    map[uint64]string{},
+		fsEval:    fsEval,
 	}
 }
 
@@ -124,7 +92,7 @@ func normalise(rawPath string, isDir bool) (string, error) {
 	// of the tar archive. While this might seem paranoid, it is a legitimate
 	// concern.
 	if "/"+path != filepath.Join("/", path) {
-		return "", errors.Errorf("escape warning: generated path is outside tar root: %s", rawPath)
+		return "", fmt.Errorf("escape warning: generated path is outside tar root: %s", rawPath)
 	}
 
 	// With some other tar formats, you needed to have a '/' at the end of a
@@ -141,24 +109,24 @@ func normalise(rawPath string, isDir bool) (string, error) {
 // the relevant stat information about the file, and also attempts to track
 // hardlinks. This should be functionally equivalent to adding entries with GNU
 // tar.
-func (tg *tarGenerator) AddFile(name, path string) error {
+func (tg *tarGenerator) AddFile(name, path string) (Err error) {
 	fi, err := tg.fsEval.Lstat(path)
 	if err != nil {
-		return errors.Wrap(err, "add file lstat")
+		return fmt.Errorf("add file lstat: %w", err)
 	}
 
 	linkname := ""
 	if fi.Mode()&os.ModeSymlink == os.ModeSymlink {
 		if linkname, err = tg.fsEval.Readlink(path); err != nil {
-			return errors.Wrap(err, "add file readlink")
+			return fmt.Errorf("add file readlink: %w", err)
 		}
 	}
 
 	hdr, err := tar.FileInfoHeader(fi, linkname)
 	if err != nil {
-		return errors.Wrap(err, "convert fi to hdr")
+		return fmt.Errorf("convert fi to hdr: %w", err)
 	}
-	hdr.Xattrs = map[string]string{}
+	hdr.Xattrs = map[string]string{} //nolint:staticcheck // SA1019: Xattrs is deprecated but PAXRecords is more annoying
 	// Usually incorrect for containers and was added in Go 1.10 causing
 	// changes to our output on a compiler bump...
 	hdr.Uname = ""
@@ -166,7 +134,7 @@ func (tg *tarGenerator) AddFile(name, path string) error {
 
 	name, err = normalise(name, fi.IsDir())
 	if err != nil {
-		return errors.Wrap(err, "normalise path")
+		return fmt.Errorf("normalise path: %w", err)
 	}
 	hdr.Name = name
 
@@ -174,7 +142,7 @@ func (tg *tarGenerator) AddFile(name, path string) error {
 	// will almost certainly confuse some users (unfortunately) but there's
 	// nothing we can do to store such files on-disk.
 	if strings.HasPrefix(filepath.Base(name), whPrefix) {
-		return errors.Errorf("invalid path has whiteout prefix %q: %s", whPrefix, name)
+		return fmt.Errorf("invalid path has whiteout prefix %q: %s", whPrefix, name)
 	}
 
 	// FIXME: Do we need to ensure that the parent paths have all been added to
@@ -190,45 +158,62 @@ func (tg *tarGenerator) AddFile(name, path string) error {
 	// by us.
 	statx, err := tg.fsEval.Lstatx(path)
 	if err != nil {
-		return errors.Wrapf(err, "lstatx %q", path)
+		return fmt.Errorf("lstatx %q: %w", path, err)
 	}
 	updateHeader(hdr, statx)
 
 	// Set up xattrs externally to updateHeader because the function signature
 	// would look really dumb otherwise.
 	// XXX: This should probably be moved to a function in tar_unix.go.
-	names, err := tg.fsEval.Llistxattr(path)
+	xattrs, err := tg.fsEval.Llistxattr(path)
 	if err != nil {
-		return errors.Wrap(err, "get xattr list")
+		if !errors.Is(err, unix.EOPNOTSUPP) {
+			return fmt.Errorf("get xattr list: %w", err)
+		}
+		xattrs = []string{}
 	}
-	for _, name := range names {
+	for _, xattr := range xattrs {
 		// Some xattrs need to be skipped for sanity reasons, such as
 		// security.selinux, because they are very much host-specific and
-		// carrying them to other hosts would be a really bad idea.
-		if _, ignore := ignoreXattrs[name]; ignore {
-			continue
+		// carrying them to other hosts would be a really bad idea. Other
+		// xattrs need to be remapped (such as escaped trusted.overlay.* xattrs
+		// when in overlayfs mode) to have correct values.
+		mappedName := xattr
+		if filter, isSpecial := getXattrFilter(xattr); isSpecial {
+			if newName := filter.ToTar(tg.onDiskFmt, xattr); newName == "" {
+				log.Debugf("xattr{%s} skipping the inclusion of xattr %q in generated tar archive", hdr.Name, xattr)
+				continue
+			} else if newName != xattr {
+				mappedName = newName
+				log.Debugf("xattr{%s} remapping xattr %q to %q in generated tar archive", hdr.Name, xattr, mappedName)
+			}
 		}
 		// TODO: We should translate all v3 capabilities into root-owned
 		//       capabilities here. But we don't have Go code for that yet
 		//       (we'd need to use libcap to parse it).
-		value, err := tg.fsEval.Lgetxattr(path, name)
+		value, err := tg.fsEval.Lgetxattr(path, xattr)
 		if err != nil {
-			// XXX: I'm not sure if we're unprivileged whether Lgetxattr can
-			//      fail with EPERM. If it can, we should ignore it (like when
-			//      we try to clear xattrs).
-			return errors.Wrapf(err, "get xattr: %s", name)
+			// Ignore xattrs we were unable to read or if the filesystem is
+			// refusing to provide information about them. Note that rather
+			// than getting a permission error when reading a trusted.* xattr
+			// as an unprivileged user, you actually get ENODATA.
+			log.Debugf("xattr{%s} failure reading xattr %q: %v", hdr.Name, xattr, err)
+			// TODO: Should we use errors.As?
+			if !errors.Is(err, unix.EOPNOTSUPP) && !errors.Is(err, unix.ENODATA) {
+				return fmt.Errorf("get xattr %q: %w", xattr, err)
+			}
 		}
 		// https://golang.org/issues/20698 -- We don't just error out here
 		// because it's not _really_ a fatal error. Currently it's unclear
 		// whether the stdlib will correctly handle reading or disable writing
 		// of these PAX headers so we have to track this ourselves.
 		if len(value) <= 0 {
-			log.Warnf("ignoring empty-valued xattr %s: disallowed by PAX standard", name)
+			log.Warnf("xattr{%s} ignoring empty-valued xattr %q: disallowed by PAX standard", hdr.Name, xattr)
 			continue
 		}
 		// Note that Go strings can actually be arbitrary byte sequences, so
 		// this conversion (while it might look a bit wrong) is actually fine.
-		hdr.Xattrs[name] = string(value)
+		hdr.Xattrs[mappedName] = string(value) //nolint:staticcheck // SA1019: Xattrs is deprecated but PAXRecords is more annoying
 	}
 
 	// Not all systems have the concept of an inode, but I'm not in the mood to
@@ -244,27 +229,27 @@ func (tg *tarGenerator) AddFile(name, path string) error {
 	}
 
 	// Apply any header mappings.
-	if err := mapHeader(hdr, tg.mapOptions); err != nil {
-		return errors.Wrap(err, "map header")
+	if err := mapHeader(hdr, tg.onDiskFmt.Map()); err != nil {
+		return fmt.Errorf("map header: %w", err)
 	}
 	if err := tg.tw.WriteHeader(hdr); err != nil {
-		return errors.Wrap(err, "write header")
+		return fmt.Errorf("write header: %w", err)
 	}
 
 	// Write the contents of regular files.
 	if hdr.Typeflag == tar.TypeReg {
 		fh, err := tg.fsEval.Open(path)
 		if err != nil {
-			return errors.Wrap(err, "open file")
+			return fmt.Errorf("open file: %w", err)
 		}
-		defer fh.Close()
+		defer funchelpers.VerifyClose(&Err, fh)
 
-		n, err := io.Copy(tg.tw, fh)
+		n, err := system.Copy(tg.tw, fh)
 		if err != nil {
-			return errors.Wrap(err, "copy to layer")
+			return fmt.Errorf("copy to layer: %w", err)
 		}
 		if n != hdr.Size {
-			return errors.Wrap(io.ErrShortWrite, "copy to layer")
+			return fmt.Errorf("copy to layer: %w", io.ErrShortWrite)
 		}
 	}
 
@@ -287,13 +272,13 @@ const whOpaque = whPrefix + whPrefix + ".opq"
 func (tg *tarGenerator) addWhiteout(name string, opaque bool) error {
 	name, err := normalise(name, false)
 	if err != nil {
-		return errors.Wrap(err, "normalise path")
+		return fmt.Errorf("normalise path: %w", err)
 	}
 
 	// Disallow having a whiteout of a whiteout, purely for our own sanity.
 	dir, file := filepath.Split(name)
 	if strings.HasPrefix(file, whPrefix) {
-		return errors.Errorf("invalid path has whiteout prefix %q: %s", whPrefix, name)
+		return fmt.Errorf("invalid path has whiteout prefix %q: %s", whPrefix, name)
 	}
 
 	// Figure out the whiteout name.
@@ -303,10 +288,10 @@ func (tg *tarGenerator) addWhiteout(name string, opaque bool) error {
 	}
 
 	// Add a dummy header for the whiteout file.
-	return errors.Wrap(tg.tw.WriteHeader(&tar.Header{
-		Name: whiteout,
-		Size: 0,
-	}), "write whiteout header")
+	if err := tg.tw.WriteHeader(&tar.Header{Name: whiteout, Size: 0}); err != nil {
+		return fmt.Errorf("write whiteout header: %w", err)
+	}
+	return nil
 }
 
 // AddWhiteout creates a whiteout for the provided path.
