@@ -1,6 +1,7 @@
+// SPDX-License-Identifier: Apache-2.0
 /*
  * umoci: Umoci Modifies Open Containers' Images
- * Copyright (C) 2016-2020 SUSE LLC
+ * Copyright (C) 2016-2025 SUSE LLC
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -15,37 +16,56 @@
  * limitations under the License.
  */
 
+// Package unpriv provides rootless emulation of CAP_DAC_READ_SEARCH without
+// the need for rootless user namespaces. This is necessary in general because
+// it turns out that a lot of distributions have a rootfs with `chmod 000`
+// directories that rely on root having CAP_DAC_READ_SEARCH to be normally
+// accessible.
+//
+// Note that the implementation of CAP_DAC_READ_SEARCH requires write access to
+// any normally-inaccessible components of paths.
+//
+// Users should use fseval.FsEval instead to allow programs to switch between
+// fseval.Rootless and fseval.Default based on whether the program is
+// privileged or not.
 package unpriv
 
 import (
 	"archive/tar"
+	"errors"
+	"fmt"
 	"os"
 	"path/filepath"
 	"sort"
 	"strings"
 	"time"
 
-	"github.com/cyphar/filepath-securejoin"
-	"github.com/opencontainers/umoci/pkg/system"
-	"github.com/pkg/errors"
+	securejoin "github.com/cyphar/filepath-securejoin"
 	"golang.org/x/sys/unix"
+
+	"github.com/opencontainers/umoci/pkg/funchelpers"
+	"github.com/opencontainers/umoci/pkg/system"
 )
 
 // fiRestore restores the state given by an os.FileInfo instance at the given
 // path by ensuring that an Lstat(path) will return as-close-to the same
 // os.FileInfo.
-//
-// #nosec G104
-func fiRestore(path string, fi os.FileInfo) {
+func fiRestore(path string, fi os.FileInfo) error {
 	// archive/tar handles the OS-specific syscall stuff required to get atime
 	// and mtime information for a file.
-	hdr, _ := tar.FileInfoHeader(fi, "")
+	hdr, err := tar.FileInfoHeader(fi, "")
+	if err != nil {
+		return err
+	}
 
 	// Apply the relevant information from the FileInfo.
-	// XXX: Should we return errors here to ensure that everything is
-	//      deterministic or we fail?
-	os.Chmod(path, fi.Mode())
-	os.Chtimes(path, hdr.AccessTime, hdr.ModTime)
+	if err := os.Chmod(path, fi.Mode()); err != nil {
+		return fmt.Errorf("restore mode: %w", err)
+	}
+	if err := os.Chtimes(path, hdr.AccessTime, hdr.ModTime); err != nil {
+		return fmt.Errorf("restore times: %w", err)
+	}
+	return nil
 }
 
 // splitpath splits the given path into each of the path components.
@@ -70,9 +90,9 @@ type WrapFunc func(path string) error
 // if the error returned is such that !os.IsPermission(err), then no trickery
 // will be performed. If fn returns an error, so will this function. All of the
 // trickery is reverted when this function returns (which is when fn returns).
-func Wrap(path string, fn WrapFunc) error {
+func Wrap(path string, fn WrapFunc) (Err error) {
 	// FIXME: Should we be calling fn() here first?
-	if err := fn(path); err == nil || !os.IsPermission(errors.Cause(err)) {
+	if err := fn(path); err == nil || !errors.Is(err, os.ErrPermission) {
 		return err
 	}
 
@@ -89,9 +109,9 @@ func Wrap(path string, fn WrapFunc) error {
 			// We've hit the first element we can chown.
 			break
 		}
-		if !os.IsPermission(err) {
+		if !errors.Is(err, os.ErrPermission) {
 			// This is a legitimate error.
-			return errors.Wrapf(err, "unpriv.wrap: lstat parent: %s", current)
+			return fmt.Errorf("unpriv.wrap: lstat parent: %s: %w", current, err)
 		}
 		start--
 	}
@@ -100,14 +120,16 @@ func Wrap(path string, fn WrapFunc) error {
 		current := filepath.Join(parts[:i]...)
 		fi, err := os.Lstat(current)
 		if err != nil {
-			return errors.Wrapf(err, "unpriv.wrap: lstat parent: %s", current)
+			return fmt.Errorf("unpriv.wrap: lstat parent: %s: %w", current, err)
 		}
 		// Add +rwx permissions to directories. If we have the access to change
 		// the mode at all then we are the user owner (not just a group owner).
-		if err := os.Chmod(current, fi.Mode()|0700); err != nil {
-			return errors.Wrapf(err, "unpriv.wrap: chmod parent: %s", current)
+		if err := os.Chmod(current, fi.Mode()|0o700); err != nil {
+			return fmt.Errorf("unpriv.wrap: chmod parent: %s: %w", current, err)
 		}
-		defer fiRestore(current, fi)
+		defer funchelpers.VerifyError(&Err, func() error {
+			return fiRestore(current, fi)
+		})
 	}
 
 	// Everything is wrapped. Return from this nightmare.
@@ -122,24 +144,31 @@ func Wrap(path string, fn WrapFunc) error {
 // doing lstat(2) may fail.
 func Open(path string) (*os.File, error) {
 	var fh *os.File
-	err := Wrap(path, func(path string) error {
+	err := Wrap(path, func(path string) (Err error) {
 		// Get information so we can revert it.
 		fi, err := os.Lstat(path)
 		if err != nil {
-			return errors.Wrap(err, "lstat file")
+			return fmt.Errorf("lstat file: %w", err)
 		}
 
-		// Add +r permissions to the file.
-		if err := os.Chmod(path, fi.Mode()|0400); err != nil {
-			return errors.Wrap(err, "chmod +r")
+		if fi.Mode()&0o400 != 0o400 {
+			// Add +r permissions to the file.
+			if err := os.Chmod(path, fi.Mode()|0o400); err != nil {
+				return fmt.Errorf("chmod +r: %w", err)
+			}
+			defer funchelpers.VerifyError(&Err, func() error {
+				return fiRestore(path, fi)
+			})
 		}
-		defer fiRestore(path, fi)
 
 		// Open the damn thing.
 		fh, err = os.Open(path)
 		return err
 	})
-	return fh, errors.Wrap(err, "unpriv.open")
+	if err != nil {
+		return nil, fmt.Errorf("unpriv.open: %w", err)
+	}
+	return fh, nil
 }
 
 // Create is a wrapper around os.Create which has been wrapped with unpriv.Wrap
@@ -154,7 +183,10 @@ func Create(path string) (*os.File, error) {
 		fh, err = os.Create(path)
 		return err
 	})
-	return fh, errors.Wrap(err, "unpriv.create")
+	if err != nil {
+		return nil, fmt.Errorf("unpriv.create: %w", err)
+	}
+	return fh, nil
 }
 
 // Readdir is a wrapper around (*os.File).Readdir which has been wrapper with
@@ -165,31 +197,36 @@ func Create(path string) (*os.File, error) {
 // resolveable).
 func Readdir(path string) ([]os.FileInfo, error) {
 	var infos []os.FileInfo
-	err := Wrap(path, func(path string) error {
+	err := Wrap(path, func(path string) (Err error) {
 		// Get information so we can revert it.
 		fi, err := os.Lstat(path)
 		if err != nil {
-			return errors.Wrap(err, "lstat dir")
+			return fmt.Errorf("lstat dir: %w", err)
 		}
 
 		// Add +rx permissions to the file.
-		if err := os.Chmod(path, fi.Mode()|0500); err != nil {
-			return errors.Wrap(err, "chmod +rx")
+		if err := os.Chmod(path, fi.Mode()|0o500); err != nil {
+			return fmt.Errorf("chmod +rx: %w", err)
 		}
-		defer fiRestore(path, fi)
+		defer funchelpers.VerifyError(&Err, func() error {
+			return fiRestore(path, fi)
+		})
 
 		// Open the damn thing.
 		fh, err := os.Open(path)
 		if err != nil {
-			return errors.Wrap(err, "opendir")
+			return fmt.Errorf("opendir: %w", err)
 		}
-		defer fh.Close()
+		defer funchelpers.VerifyClose(&Err, fh)
 
 		// Get the set of dirents.
 		infos, err = fh.Readdir(-1)
 		return err
 	})
-	return infos, errors.Wrap(err, "unpriv.readdir")
+	if err != nil {
+		return nil, fmt.Errorf("unpriv.readdir: %w", err)
+	}
+	return infos, nil
 }
 
 // Lstat is a wrapper around os.Lstat which has been wrapped with unpriv.Wrap
@@ -205,16 +242,22 @@ func Lstat(path string) (os.FileInfo, error) {
 		fi, err = os.Lstat(path)
 		return err
 	})
-	return fi, errors.Wrap(err, "unpriv.lstat")
+	if err != nil {
+		return nil, fmt.Errorf("unpriv.lstat: %w", err)
+	}
+	return fi, nil
 }
 
-// Lstatx is like Lstat but uses unix.Lstat and returns unix.Stat_t instead
+// Lstatx is like Lstat but uses unix.Lstat and returns unix.Stat_t instead.
 func Lstatx(path string) (unix.Stat_t, error) {
 	var s unix.Stat_t
 	err := Wrap(path, func(path string) error {
 		return unix.Lstat(path, &s)
 	})
-	return s, errors.Wrap(err, "unpriv.lstatx")
+	if err != nil {
+		return s, fmt.Errorf("unpriv.lstatx: %w", err)
+	}
+	return s, nil
 }
 
 // Readlink is a wrapper around os.Readlink which has been wrapped with
@@ -230,7 +273,10 @@ func Readlink(path string) (string, error) {
 		target, err = os.Readlink(path)
 		return err
 	})
-	return target, errors.Wrap(err, "unpriv.readlink")
+	if err != nil {
+		return "", fmt.Errorf("unpriv.readlink: %w", err)
+	}
+	return target, nil
 }
 
 // Symlink is a wrapper around os.Symlink which has been wrapped with
@@ -239,9 +285,11 @@ func Readlink(path string) (string, error) {
 // may not have resolve access after this function returns because all of the
 // trickery is reverted by unpriv.Wrap.
 func Symlink(target, linkname string) error {
-	return errors.Wrap(Wrap(linkname, func(linkname string) error {
-		return os.Symlink(target, linkname)
-	}), "unpriv.symlink")
+	err := Wrap(linkname, func(linkname string) error { return os.Symlink(target, linkname) })
+	if err != nil {
+		return fmt.Errorf("unpriv.symlink: %w", err)
+	}
+	return nil
 }
 
 // Link is a wrapper around unix.Link(..., 0) which has been wrapped with
@@ -250,65 +298,71 @@ func Symlink(target, linkname string) error {
 // you may not have resolve access after this function returns because all of
 // the trickery is reverted by unpriv.Wrap.
 func Link(target, linkname string) error {
-	return errors.Wrap(Wrap(linkname, func(linkname string) error {
+	err := Wrap(linkname, func(linkname string) error {
 		// We have to double-wrap this, because you need search access to the
 		// linkname. This is safe because any common ancestors will be reverted
 		// in reverse call stack order.
-		return errors.Wrap(Wrap(target, func(target string) error {
+		err := Wrap(target, func(target string) error {
 			// We need to explicitly pass 0 as a flag because POSIX allows the
 			// default behaviour of link(2) when it comes to target being a
 			// symlink to be implementation-defined. Only linkat(2) allows us
 			// to guarantee the right behaviour.
 			//  <https://pubs.opengroup.org/onlinepubs/9699919799/functions/link.html>
 			return unix.Linkat(unix.AT_FDCWD, target, unix.AT_FDCWD, linkname, 0)
-		}), "unpriv.wrap target")
-	}), "unpriv.link")
+		})
+		if err != nil {
+			return fmt.Errorf("unpriv.wrap target: %w", err)
+		}
+		return nil
+	})
+	if err != nil {
+		return fmt.Errorf("unpriv.link: %w", err)
+	}
+	return nil
 }
 
 // Chmod is a wrapper around os.Chmod which has been wrapped with unpriv.Wrap
 // to make it possible to change the permission bits of a path even if you do
 // not currently have the required access bits to access the path.
 func Chmod(path string, mode os.FileMode) error {
-	return errors.Wrap(Wrap(path, func(path string) error {
-		return os.Chmod(path, mode)
-	}), "unpriv.chmod")
-}
-
-// Lchown is a wrapper around os.Lchown which has been wrapped with unpriv.Wrap
-// to make it possible to change the owner of a path even if you do not
-// currently have the required access bits to access the path. Note that this
-// function is not particularly useful in most rootless scenarios.
-//
-// FIXME: This probably should be removed because it's questionably useful.
-func Lchown(path string, uid, gid int) error {
-	return errors.Wrap(Wrap(path, func(path string) error {
-		return os.Lchown(path, uid, gid)
-	}), "unpriv.lchown")
+	err := Wrap(path, func(path string) error { return os.Chmod(path, mode) })
+	if err != nil {
+		return fmt.Errorf("unpriv.chmod: %w", err)
+	}
+	return nil
 }
 
 // Chtimes is a wrapper around os.Chtimes which has been wrapped with
 // unpriv.Wrap to make it possible to change the modified times of a path even
 // if you do not currently have the required access bits to access the path.
 func Chtimes(path string, atime, mtime time.Time) error {
-	return errors.Wrap(Wrap(path, func(path string) error {
-		return os.Chtimes(path, atime, mtime)
-	}), "unpriv.chtimes")
+	err := Wrap(path, func(path string) error { return os.Chtimes(path, atime, mtime) })
+	if err != nil {
+		return fmt.Errorf("unpriv.chtimes: %w", err)
+	}
+	return nil
 }
 
 // Lutimes is a wrapper around system.Lutimes which has been wrapped with
 // unpriv.Wrap to make it possible to change the modified times of a path even
 // if you do no currently have the required access bits to access the path.
 func Lutimes(path string, atime, mtime time.Time) error {
-	return errors.Wrap(Wrap(path, func(path string) error {
-		return system.Lutimes(path, atime, mtime)
-	}), "unpriv.lutimes")
+	err := Wrap(path, func(path string) error { return system.Lutimes(path, atime, mtime) })
+	if err != nil {
+		return fmt.Errorf("unpriv.lutimes: %w", err)
+	}
+	return nil
 }
 
 // Remove is a wrapper around os.Remove which has been wrapped with unpriv.Wrap
 // to make it possible to remove a path even if you do not currently have the
 // required access bits to modify or resolve the path.
 func Remove(path string) error {
-	return errors.Wrap(Wrap(path, os.Remove), "unpriv.remove")
+	err := Wrap(path, os.Remove)
+	if err != nil {
+		return fmt.Errorf("unpriv.remove: %w", err)
+	}
+	return nil
 }
 
 // foreachSubpath executes WrapFunc for each child of the given path (not
@@ -316,11 +370,11 @@ func Remove(path string) error {
 // not be called and no error will be returned. This should be called within a
 // context where path has already been made resolveable, however the . If WrapFunc returns an
 // error, the first error is returned and iteration is halted.
-func foreachSubpath(path string, wrapFn WrapFunc) error {
+func foreachSubpath(path string, wrapFn WrapFunc) (Err error) {
 	// Is the path a directory?
 	fi, err := os.Lstat(path)
 	if err != nil {
-		return errors.WithStack(err)
+		return err
 	}
 	if !fi.IsDir() {
 		return nil
@@ -329,20 +383,24 @@ func foreachSubpath(path string, wrapFn WrapFunc) error {
 	// Open the directory.
 	fd, err := Open(path)
 	if err != nil {
-		return errors.WithStack(err)
+		return err
 	}
-	defer fd.Close()
+	defer funchelpers.VerifyClose(&Err, fd)
 
 	// We need to change the mode to Readdirnames. We don't need to worry about
 	// permissions because we're already in a context with filepath.Dir(path)
 	// is at least a+rx. However, because we are calling wrapFn we need to
 	// restore the original mode immediately.
-	// #nosec G104
-	_ = os.Chmod(path, fi.Mode()|0444)
+	if err := os.Chmod(path, fi.Mode()|0o444); err != nil {
+		return fmt.Errorf("chmod +r to readdir: %w", err)
+	}
+	defer funchelpers.VerifyError(&Err, func() error {
+		return fiRestore(path, fi)
+	})
+
 	names, err := fd.Readdirnames(-1)
-	fiRestore(path, fi)
 	if err != nil {
-		return errors.WithStack(err)
+		return err
 	}
 
 	// Make iteration order consistent.
@@ -363,10 +421,10 @@ func foreachSubpath(path string, wrapFn WrapFunc) error {
 // wrapped with unpriv.Wrap to make it possible to remove a path (even if it
 // has child paths) even if you do not currently have enough access bits.
 func RemoveAll(path string) error {
-	return errors.Wrap(Wrap(path, func(path string) error {
+	err := Wrap(path, func(path string) error {
 		// If remove works, we're done.
 		err := os.Remove(path)
-		if err == nil || os.IsNotExist(errors.Cause(err)) {
+		if err == nil || errors.Is(err, os.ErrNotExist) {
 			return nil
 		}
 
@@ -374,17 +432,16 @@ func RemoveAll(path string) error {
 		fi, serr := os.Lstat(path)
 		if serr != nil {
 			// Use securejoin's IsNotExist to handle ENOTDIR sanely.
-			if securejoin.IsNotExist(errors.Cause(serr)) {
-				serr = nil
+			if !securejoin.IsNotExist(serr) {
+				return fmt.Errorf("lstat: %w", serr)
 			}
-			return errors.Wrap(serr, "lstat")
+			return nil
 		}
 		// Return error from remove if it's not a directory.
 		if !fi.IsDir() {
-			return errors.Wrap(err, "remove non-directory")
+			return fmt.Errorf("remove non-directory: %w", err)
 		}
 		err = nil
-
 		err1 := foreachSubpath(path, func(subpath string) error {
 			err2 := RemoveAll(subpath)
 			if err == nil {
@@ -394,38 +451,44 @@ func RemoveAll(path string) error {
 		})
 		if err1 != nil {
 			// We must have hit a race, but we don't care.
-			if os.IsNotExist(errors.Cause(err1)) {
+			if errors.Is(err1, os.ErrNotExist) {
 				err1 = nil
 			}
-			return errors.Wrap(err1, "foreach subpath")
+			return fmt.Errorf("foreach subpath: %w", err1)
 		}
 
 		// Remove the directory. This should now work.
 		err1 = os.Remove(path)
-		if err1 == nil || os.IsNotExist(errors.Cause(err1)) {
+		if err1 == nil || errors.Is(err1, os.ErrNotExist) {
 			return nil
 		}
 		if err == nil {
 			err = err1
 		}
-		return errors.Wrap(err, "remove")
-	}), "unpriv.removeall")
+		return fmt.Errorf("remove: %w", err)
+	})
+	if err != nil {
+		return fmt.Errorf("unpriv.removeall: %w", err)
+	}
+	return nil
 }
 
 // Mkdir is a wrapper around os.Mkdir which has been wrapped with unpriv.Wrap
 // to make it possible to remove a path even if you do not currently have the
 // required access bits to modify or resolve the path.
 func Mkdir(path string, perm os.FileMode) error {
-	return errors.Wrap(Wrap(path, func(path string) error {
-		return os.Mkdir(path, perm)
-	}), "unpriv.mkdir")
+	err := Wrap(path, func(path string) error { return os.Mkdir(path, perm) })
+	if err != nil {
+		return fmt.Errorf("unpriv.mkdir: %w", err)
+	}
+	return nil
 }
 
 // MkdirAll is similar to os.MkdirAll but in order to implement it properly all
 // of the internal functions were wrapped with unpriv.Wrap to make it possible
 // to create a path even if you do not currently have enough access bits.
 func MkdirAll(path string, perm os.FileMode) error {
-	return errors.Wrap(Wrap(path, func(path string) error {
+	err := Wrap(path, func(path string) error {
 		// Check whether the path already exists.
 		fi, err := os.Stat(path)
 		if err == nil {
@@ -455,16 +518,22 @@ func MkdirAll(path string, perm os.FileMode) error {
 			return err
 		}
 		return nil
-	}), "unpriv.mkdirall")
+	})
+	if err != nil {
+		return fmt.Errorf("unpriv.mkdirall: %w", err)
+	}
+	return nil
 }
 
 // Mknod is a wrapper around unix.Mknod which has been wrapped with unpriv.Wrap
 // to make it possible to remove a path even if you do not currently have the
 // required access bits to modify or resolve the path.
 func Mknod(path string, mode os.FileMode, dev uint64) error {
-	return errors.Wrap(Wrap(path, func(path string) error {
-		return unix.Mknod(path, uint32(mode), int(dev))
-	}), "unpriv.mknod")
+	err := Wrap(path, func(path string) error { return system.Mknod(path, uint32(mode), dev) })
+	if err != nil {
+		return fmt.Errorf("unpriv.mknod: %w", err)
+	}
+	return nil
 }
 
 // Llistxattr is a wrapper around system.Llistxattr which has been wrapped with
@@ -477,25 +546,32 @@ func Llistxattr(path string) ([]string, error) {
 		xattrs, err = system.Llistxattr(path)
 		return err
 	})
-	return xattrs, errors.Wrap(err, "unpriv.llistxattr")
+	if err != nil {
+		return nil, fmt.Errorf("unpriv.llistxattr: %w", err)
+	}
+	return xattrs, nil
 }
 
 // Lremovexattr is a wrapper around system.Lremovexattr which has been wrapped
 // with unpriv.Wrap to make it possible to remove a path even if you do not
 // currently have the required access bits to resolve the path.
 func Lremovexattr(path, name string) error {
-	return errors.Wrap(Wrap(path, func(path string) error {
-		return unix.Lremovexattr(path, name)
-	}), "unpriv.lremovexattr")
+	err := Wrap(path, func(path string) error { return unix.Lremovexattr(path, name) })
+	if err != nil {
+		return fmt.Errorf("unpriv.lremovexattr: %w", err)
+	}
+	return nil
 }
 
 // Lsetxattr is a wrapper around system.Lsetxattr which has been wrapped
 // with unpriv.Wrap to make it possible to set a path even if you do not
 // currently have the required access bits to resolve the path.
 func Lsetxattr(path, name string, value []byte, flags int) error {
-	return errors.Wrap(Wrap(path, func(path string) error {
-		return unix.Lsetxattr(path, name, value, flags)
-	}), "unpriv.lsetxattr")
+	err := Wrap(path, func(path string) error { return unix.Lsetxattr(path, name, value, flags) })
+	if err != nil {
+		return fmt.Errorf("unpriv.lsetxattr: %w", err)
+	}
+	return nil
 }
 
 // Lgetxattr is a wrapper around system.Lgetxattr which has been wrapped
@@ -508,36 +584,21 @@ func Lgetxattr(path, name string) ([]byte, error) {
 		value, err = system.Lgetxattr(path, name)
 		return err
 	})
-	return value, errors.Wrap(err, "unpriv.lgetxattr")
+	if err != nil {
+		return nil, fmt.Errorf("unpriv.lgetxattr: %w", err)
+	}
+	return value, nil
 }
 
-// Lclearxattrs is similar to system.Lclearxattrs but in order to implement it
-// properly all of the internal functions were wrapped with unpriv.Wrap to make
-// it possible to create a path even if you do not currently have enough access
-// bits.
-func Lclearxattrs(path string, except map[string]struct{}) error {
-	return errors.Wrap(Wrap(path, func(path string) error {
-		names, err := Llistxattr(path)
-		if err != nil {
-			return err
-		}
-		for _, name := range names {
-			if _, skip := except[name]; skip {
-				continue
-			}
-			if err := Lremovexattr(path, name); err != nil {
-				// SELinux won't let you change security.selinux (for obvious
-				// security reasons), so we don't clear xattrs if attempting to
-				// clear them causes an EPERM. This EPERM will not be due to
-				// resolution issues (Llistxattr already has done that for us).
-				if os.IsPermission(errors.Cause(err)) {
-					continue
-				}
-				return err
-			}
-		}
-		return nil
-	}), "unpriv.lclearxattrs")
+// Lclearxattrs is a wrapper around system.Lclearxattrs which has been wrapped
+// with unpriv.Wrap to make it possible to get a path even if you do not
+// currently have the required access bits to resolve the path.
+func Lclearxattrs(path string, skipFn func(xattrName string) bool) error {
+	err := Wrap(path, func(path string) error { return system.Lclearxattrs(path, skipFn) })
+	if err != nil {
+		return fmt.Errorf("unpriv.lclearxattrs: %w", err)
+	}
+	return nil
 }
 
 // walk is the inner implementation of Walk.
@@ -550,20 +611,25 @@ func walk(path string, info os.FileInfo, walkFn filepath.WalkFunc) error {
 	}
 
 	// Now just execute walkFn over each subpath.
+	// TODO: We should handle the Readdirnames failing case that stdlib does.
 	return foreachSubpath(path, func(subpath string) error {
 		info, err := Lstat(subpath)
 		if err != nil {
 			// If it doesn't exist, just pass it directly to walkFn.
 			if err := walkFn(subpath, info, err); err != nil {
-				// Ignore SkipDir.
-				if errors.Cause(err) != filepath.SkipDir {
+				// To match stdlib, SkipDir assumes a non-existent path is a
+				// directory and so SkipDir just skips that path.
+				if errors.Is(err, filepath.SkipDir) {
 					return err
 				}
 			}
 		} else {
 			if err := walk(subpath, info, walkFn); err != nil {
-				// Ignore error if it's SkipDir and subpath is a directory.
-				if !(info.IsDir() && errors.Cause(err) == filepath.SkipDir) {
+				// If this entry is a directory then SkipDir will just skip
+				// this entry and continue walking the current directory, but
+				// otherwise we need to skip the whole directory. This matches
+				// the stdlib behaviour.
+				if !(info.IsDir() && errors.Is(err, filepath.SkipDir)) { //nolint:staticcheck // QF1001: this form is easier to understand
 					return err
 				}
 			}
@@ -585,9 +651,9 @@ func Walk(root string, walkFn filepath.WalkFunc) error {
 		} else {
 			err = walk(root, info, walkFn)
 		}
-		if errors.Cause(err) == filepath.SkipDir {
-			err = nil
+		if err == nil || errors.Is(err, filepath.SkipDir) {
+			return nil
 		}
-		return errors.Wrap(err, "unpriv.walk")
+		return fmt.Errorf("unpriv.walk: %w", err)
 	})
 }
