@@ -1,6 +1,7 @@
+// SPDX-License-Identifier: Apache-2.0
 /*
  * umoci: Umoci Modifies Open Containers' Images
- * Copyright (C) 2016-2020 SUSE LLC
+ * Copyright (C) 2016-2025 SUSE LLC
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -19,24 +20,35 @@ package umoci
 
 import (
 	"context"
+	"fmt"
 	"os"
 	"path/filepath"
 	"strings"
 
 	"github.com/apex/log"
 	ispec "github.com/opencontainers/image-spec/specs-go/v1"
+	"github.com/vbatts/go-mtree"
+
 	"github.com/opencontainers/umoci/mutate"
 	"github.com/opencontainers/umoci/oci/casext"
 	"github.com/opencontainers/umoci/oci/layer"
 	"github.com/opencontainers/umoci/pkg/fseval"
+	"github.com/opencontainers/umoci/pkg/funchelpers"
 	"github.com/opencontainers/umoci/pkg/mtreefilter"
-	"github.com/pkg/errors"
-	"github.com/vbatts/go-mtree"
 )
 
 // Repack repacks a bundle into an image adding a new layer for the changed
 // data in the bundle.
-func Repack(engineExt casext.Engine, tagName string, bundlePath string, meta Meta, history *ispec.History, filters []mtreefilter.FilterFunc, refreshBundle bool, mutator *mutate.Mutator) error {
+//
+// If layerCompressor is nil, the compression algorithm is auto-selected.
+func Repack(engineExt casext.Engine, tagName, bundlePath string,
+	meta Meta,
+	history *ispec.History,
+	filters []mtreefilter.FilterFunc,
+	refreshBundle bool,
+	mutator *mutate.Mutator,
+	layerCompressor mutate.Compressor, //nolint:staticcheck // SA1019: this interface is defined by us and we keep it for compatibility
+) (Err error) {
 	mtreeName := strings.Replace(meta.From.Descriptor().Digest.String(), ":", "_", 1)
 	mtreePath := filepath.Join(bundlePath, mtreeName+".mtree")
 	fullRootfsPath := filepath.Join(bundlePath, layer.RootfsName)
@@ -49,13 +61,13 @@ func Repack(engineExt casext.Engine, tagName string, bundlePath string, meta Met
 
 	mfh, err := os.Open(mtreePath)
 	if err != nil {
-		return errors.Wrap(err, "open mtree")
+		return fmt.Errorf("open mtree: %w", err)
 	}
-	defer mfh.Close()
+	defer funchelpers.VerifyClose(&Err, mfh)
 
 	spec, err := mtree.ParseSpec(mfh)
 	if err != nil {
-		return errors.Wrap(err, "parse mtree")
+		return fmt.Errorf("parse mtree: %w", err)
 	}
 
 	log.WithFields(log.Fields{
@@ -70,7 +82,7 @@ func Repack(engineExt casext.Engine, tagName string, bundlePath string, meta Met
 	log.Info("computing filesystem diff ...")
 	diffs, err := mtree.Check(fullRootfsPath, spec, MtreeKeywords, fsEval)
 	if err != nil {
-		return errors.Wrap(err, "check mtree")
+		return fmt.Errorf("check mtree: %w", err)
 	}
 	log.Info("... done")
 
@@ -97,37 +109,35 @@ func Repack(engineExt casext.Engine, tagName string, bundlePath string, meta Met
 			return err
 		}
 
-		err = mutator.Set(context.Background(), config, imageMeta, annotations, history)
+		err = mutator.Set(context.Background(), config.Config, imageMeta, annotations, history)
 		if err != nil {
 			return err
 		}
 	} else {
-		packOptions := layer.RepackOptions{MapOptions: meta.MapOptions}
-		if meta.WhiteoutMode == layer.OverlayFSWhiteout {
-			packOptions.TranslateOverlayWhiteouts = true
-		}
-		reader, err := layer.GenerateLayer(fullRootfsPath, diffs, &packOptions)
+		reader, err := layer.GenerateLayer(fullRootfsPath, diffs, &layer.RepackOptions{
+			OnDiskFormat: layer.DirRootfs{
+				MapOptions: meta.MapOptions,
+			},
+		})
 		if err != nil {
-			return errors.Wrap(err, "generate diff layer")
+			return fmt.Errorf("generate diff layer: %w", err)
 		}
-		defer reader.Close()
+		defer funchelpers.VerifyClose(&Err, reader)
 
-		// TODO: We should add a flag to allow for a new layer to be made
-		//       non-distributable.
-		if _, err := mutator.Add(context.Background(), ispec.MediaTypeImageLayer, reader, history, mutate.GzipCompressor); err != nil {
-			return errors.Wrap(err, "add diff layer")
+		if _, err := mutator.Add(context.Background(), ispec.MediaTypeImageLayer, reader, history, layerCompressor, nil); err != nil {
+			return fmt.Errorf("add diff layer: %w", err)
 		}
 	}
 
 	newDescriptorPath, err := mutator.Commit(context.Background())
 	if err != nil {
-		return errors.Wrap(err, "commit mutated image")
+		return fmt.Errorf("commit mutated image: %w", err)
 	}
 
 	log.Infof("new image manifest created: %s->%s", newDescriptorPath.Root().Digest, newDescriptorPath.Descriptor().Digest)
 
 	if err := engineExt.UpdateReference(context.Background(), tagName, newDescriptorPath.Root()); err != nil {
-		return errors.Wrap(err, "add new tag")
+		return fmt.Errorf("add new tag: %w", err)
 	}
 
 	log.Infof("created new tag for image manifest: %s", tagName)
@@ -135,14 +145,14 @@ func Repack(engineExt casext.Engine, tagName string, bundlePath string, meta Met
 	if refreshBundle {
 		newMtreeName := strings.Replace(newDescriptorPath.Descriptor().Digest.String(), ":", "_", 1)
 		if err := GenerateBundleManifest(newMtreeName, bundlePath, fsEval); err != nil {
-			return errors.Wrap(err, "write mtree metadata")
+			return fmt.Errorf("write mtree metadata: %w", err)
 		}
 		if err := os.Remove(mtreePath); err != nil {
-			return errors.Wrap(err, "remove old mtree metadata")
+			return fmt.Errorf("remove old mtree metadata: %w", err)
 		}
 		meta.From = newDescriptorPath
 		if err := WriteBundleMeta(bundlePath, meta); err != nil {
-			return errors.Wrap(err, "write umoci.json metadata")
+			return fmt.Errorf("write umoci.json metadata: %w", err)
 		}
 	}
 	return nil
