@@ -1,6 +1,7 @@
+// SPDX-License-Identifier: Apache-2.0
 /*
  * umoci: Umoci Modifies Open Containers' Images
- * Copyright (C) 2016-2020 SUSE LLC
+ * Copyright (C) 2016-2025 SUSE LLC
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -25,40 +26,41 @@ import (
 
 	"github.com/apex/log"
 	ispec "github.com/opencontainers/image-spec/specs-go/v1"
+
 	"github.com/opencontainers/umoci/oci/casext"
 	"github.com/opencontainers/umoci/oci/layer"
 	"github.com/opencontainers/umoci/pkg/fseval"
-	"github.com/pkg/errors"
+	"github.com/opencontainers/umoci/pkg/funchelpers"
 )
 
 // Unpack unpacks an image to the specified bundle path.
-func Unpack(engineExt casext.Engine, fromName string, bundlePath string, unpackOptions layer.UnpackOptions) error {
-	var meta Meta
-	meta.Version = MetaVersion
-	meta.MapOptions = unpackOptions.MapOptions
-	meta.WhiteoutMode = unpackOptions.WhiteoutMode
+func Unpack(engineExt casext.Engine, fromName, bundlePath string, unpackOptions layer.UnpackOptions) (Err error) {
+	meta := Meta{
+		Version:    MetaVersion,
+		MapOptions: unpackOptions.MapOptions(),
+	}
 
 	fromDescriptorPaths, err := engineExt.ResolveReference(context.Background(), fromName)
 	if err != nil {
-		return errors.Wrap(err, "get descriptor")
+		return fmt.Errorf("get descriptor: %w", err)
 	}
 	if len(fromDescriptorPaths) == 0 {
-		return errors.Errorf("tag is not found: %s", fromName)
+		return fmt.Errorf("tag is not found: %s", fromName)
 	}
 	if len(fromDescriptorPaths) != 1 {
 		// TODO: Handle this more nicely.
-		return errors.Errorf("tag is ambiguous: %s", fromName)
+		return fmt.Errorf("tag is ambiguous: %s", fromName)
 	}
 	meta.From = fromDescriptorPaths[0]
 
 	manifestBlob, err := engineExt.FromDescriptor(context.Background(), meta.From.Descriptor())
 	if err != nil {
-		return errors.Wrap(err, "get manifest")
+		return fmt.Errorf("get manifest: %w", err)
 	}
-	defer manifestBlob.Close()
+	defer funchelpers.VerifyClose(&Err, manifestBlob)
 
 	if manifestBlob.Descriptor.MediaType != ispec.MediaTypeImageManifest {
-		return errors.Wrap(fmt.Errorf("descriptor does not point to ispec.MediaTypeImageManifest: not implemented: %s", manifestBlob.Descriptor.MediaType), "invalid --image tag")
+		return fmt.Errorf("invalid --image tag: descriptor does not point to ispec.MediaTypeImageManifest: not implemented: %s", manifestBlob.Descriptor.MediaType)
 	}
 
 	mtreeName := strings.Replace(meta.From.Descriptor().Digest.String(), ":", "_", 1)
@@ -72,18 +74,18 @@ func Unpack(engineExt casext.Engine, fromName string, bundlePath string, unpackO
 	manifest, ok := manifestBlob.Data.(ispec.Manifest)
 	if !ok {
 		// Should _never_ be reached.
-		return errors.Errorf("[internal error] unknown manifest blob type: %s", manifestBlob.Descriptor.MediaType)
+		return fmt.Errorf("[internal error] unknown manifest blob type: %s", manifestBlob.Descriptor.MediaType)
 	}
 
 	// Unpack the runtime bundle.
-	if err := os.MkdirAll(bundlePath, 0755); err != nil {
-		return errors.Wrap(err, "create bundle path")
+	if err := os.MkdirAll(bundlePath, 0o755); err != nil {
+		return fmt.Errorf("create bundle path: %w", err)
 	}
 	// XXX: We should probably defer os.RemoveAll(bundlePath).
 
 	log.Info("unpacking bundle ...")
 	if err := layer.UnpackManifest(context.Background(), engineExt, bundlePath, manifest, &unpackOptions); err != nil {
-		return errors.Wrap(err, "create runtime bundle")
+		return fmt.Errorf("create runtime bundle: %w", err)
 	}
 	log.Info("... done")
 
@@ -93,7 +95,7 @@ func Unpack(engineExt casext.Engine, fromName string, bundlePath string, unpackO
 	}
 
 	if err := GenerateBundleManifest(mtreeName, bundlePath, fsEval); err != nil {
-		return errors.Wrap(err, "write mtree")
+		return fmt.Errorf("write mtree: %w", err)
 	}
 
 	log.WithFields(log.Fields{
@@ -103,7 +105,7 @@ func Unpack(engineExt casext.Engine, fromName string, bundlePath string, unpackO
 	}).Debugf("umoci: saving Meta metadata")
 
 	if err := WriteBundleMeta(bundlePath, meta); err != nil {
-		return errors.Wrap(err, "write umoci.json metadata")
+		return fmt.Errorf("write umoci.json metadata: %w", err)
 	}
 
 	log.Infof("unpacked image bundle: %s", bundlePath)
