@@ -1,6 +1,7 @@
+// SPDX-License-Identifier: Apache-2.0
 /*
  * umoci: Umoci Modifies Open Containers' Images
- * Copyright (C) 2016-2020 SUSE LLC
+ * Copyright (C) 2016-2025 SUSE LLC
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -31,13 +32,14 @@ import (
 	"github.com/apex/log"
 	"github.com/docker/go-units"
 	ispec "github.com/opencontainers/image-spec/specs-go/v1"
+	"github.com/urfave/cli"
+	"github.com/vbatts/go-mtree"
+
 	"github.com/opencontainers/umoci/oci/casext"
 	igen "github.com/opencontainers/umoci/oci/config/generate"
 	"github.com/opencontainers/umoci/oci/layer"
+	"github.com/opencontainers/umoci/pkg/funchelpers"
 	"github.com/opencontainers/umoci/pkg/idtools"
-	"github.com/pkg/errors"
-	"github.com/urfave/cli"
-	"github.com/vbatts/go-mtree"
 )
 
 // FIXME: This should be moved to a library. Too much of this code is in the
@@ -92,7 +94,10 @@ type Meta struct {
 
 	// WhiteoutMode indicates what style of whiteout was written to disk
 	// when this filesystem was extracted.
-	WhiteoutMode layer.WhiteoutMode `json:"whiteout_mode"`
+	// NOTE: This field has been deprecated, as the feature was completely
+	// broken. See <https://github.com/opencontainers/umoci/issues/574> for
+	// more details.
+	DeprecatedWhiteoutMode int `json:"whiteout_mode,omitempty"`
 }
 
 // WriteTo writes a JSON-serialised version of Meta to the given io.Writer.
@@ -103,39 +108,49 @@ func (m Meta) WriteTo(w io.Writer) (int64, error) {
 }
 
 // WriteBundleMeta writes an umoci.json file to the given bundle path.
-func WriteBundleMeta(bundle string, meta Meta) error {
+func WriteBundleMeta(bundle string, meta Meta) (Err error) {
 	fh, err := os.Create(filepath.Join(bundle, MetaName))
 	if err != nil {
-		return errors.Wrap(err, "create metadata")
+		return fmt.Errorf("create metadata: %w", err)
 	}
-	defer fh.Close()
+	defer funchelpers.VerifyClose(&Err, fh)
 
-	_, err = meta.WriteTo(fh)
-	return errors.Wrap(err, "write metadata")
+	if _, err := meta.WriteTo(fh); err != nil {
+		return fmt.Errorf("write metadata: %w", err)
+	}
+	return nil
 }
 
 // ReadBundleMeta reads and parses the umoci.json file from a given bundle path.
-func ReadBundleMeta(bundle string) (Meta, error) {
+func ReadBundleMeta(bundle string) (_ Meta, Err error) {
 	var meta Meta
 
 	fh, err := os.Open(filepath.Join(bundle, MetaName))
 	if err != nil {
-		return meta, errors.Wrap(err, "open metadata")
+		return meta, fmt.Errorf("open metadata: %w", err)
 	}
-	defer fh.Close()
+	defer funchelpers.VerifyClose(&Err, fh)
 
 	err = json.NewDecoder(fh).Decode(&meta)
-	if meta.Version != MetaVersion {
-		if err == nil {
-			err = fmt.Errorf("unsupported umoci.json version: %s", meta.Version)
-		}
+	if err != nil {
+		return meta, fmt.Errorf("decode metadata: %w", err)
 	}
-	return meta, errors.Wrap(err, "decode metadata")
+	if meta.Version != MetaVersion {
+		return meta, fmt.Errorf("decode metadata: unsupported umoci.json version %q", meta.Version)
+	}
+	// NOTE: This field has been deprecated, as the feature was completely
+	// broken. See <https://github.com/opencontainers/umoci/issues/574> for
+	// more details.
+	if meta.DeprecatedWhiteoutMode != 0 {
+		return meta, fmt.Errorf("decode metadata: deprecated (broken) whiteout_mode field set (%d)", meta.DeprecatedWhiteoutMode)
+	}
+	return meta, nil
 }
 
 // ManifestStat has information about a given OCI manifest.
 // TODO: Implement support for manifest lists, this should also be able to
-//       contain stat information for a list of manifests.
+//
+//	contain stat information for a list of manifests.
 type ManifestStat struct {
 	// TODO: Flesh this out. Currently it's only really being used to get an
 	//       equivalent of docker-history(1). We really need to add more
@@ -148,17 +163,20 @@ type ManifestStat struct {
 // Format formats a ManifestStat using the default formatting, and writes the
 // result to the given writer.
 // TODO: This should really be implemented in a way that allows for users to
-//       define their own custom templates for different blocks (meaning that
-//       this should use text/template rather than using tabwriters manually.
+//
+//	define their own custom templates for different blocks (meaning that
+//	this should use text/template rather than using tabwriters manually.
 func (ms ManifestStat) Format(w io.Writer) error {
 	// Output history information.
 	tw := tabwriter.NewWriter(w, 4, 2, 1, ' ', 0)
-	fmt.Fprintf(tw, "LAYER\tCREATED\tCREATED BY\tSIZE\tCOMMENT\n")
+	if _, err := fmt.Fprintf(tw, "LAYER\tCREATED\tCREATED BY\tSIZE\tCOMMENT\n"); err != nil {
+		return err
+	}
 	for _, histEntry := range ms.History {
 		var (
-			created   = strings.Replace(histEntry.Created.Format(igen.ISO8601), "\t", " ", -1)
-			createdBy = strings.Replace(histEntry.CreatedBy, "\t", " ", -1)
-			comment   = strings.Replace(histEntry.Comment, "\t", " ", -1)
+			created   = strings.ReplaceAll(histEntry.Created.Format(igen.ISO8601), "\t", " ")
+			createdBy = strings.ReplaceAll(histEntry.CreatedBy, "\t", " ")
+			comment   = strings.ReplaceAll(histEntry.Comment, "\t", " ")
 			layerID   = "<none>"
 			size      = "<none>"
 		)
@@ -169,7 +187,9 @@ func (ms ManifestStat) Format(w io.Writer) error {
 		}
 
 		// TODO: We need to truncate some of the fields.
-		fmt.Fprintf(tw, "%s\t%s\t%s\t%s\t%s\n", layerID, created, createdBy, size, comment)
+		if _, err := fmt.Fprintf(tw, "%s\t%s\t%s\t%s\t%s\n", layerID, created, createdBy, size, comment); err != nil {
+			return err
+		}
 	}
 	return tw.Flush()
 }
@@ -198,7 +218,7 @@ func Stat(ctx context.Context, engine casext.Engine, manifestDescriptor ispec.De
 	var stat ManifestStat
 
 	if manifestDescriptor.MediaType != ispec.MediaTypeImageManifest {
-		return stat, errors.Errorf("stat: cannot stat a non-manifest descriptor: invalid media type '%s'", manifestDescriptor.MediaType)
+		return stat, fmt.Errorf("stat: cannot stat a non-manifest descriptor: invalid media type %q", manifestDescriptor.MediaType)
 	}
 
 	// We have to get the actual manifest.
@@ -209,18 +229,18 @@ func Stat(ctx context.Context, engine casext.Engine, manifestDescriptor ispec.De
 	manifest, ok := manifestBlob.Data.(ispec.Manifest)
 	if !ok {
 		// Should _never_ be reached.
-		return stat, errors.Errorf("[internal error] unknown manifest blob type: %s", manifestBlob.Descriptor.MediaType)
+		return stat, fmt.Errorf("[internal error] unknown manifest blob type: %s", manifestBlob.Descriptor.MediaType)
 	}
 
 	// Now get the config.
 	configBlob, err := engine.FromDescriptor(ctx, manifest.Config)
 	if err != nil {
-		return stat, errors.Wrap(err, "stat")
+		return stat, fmt.Errorf("stat: %w", err)
 	}
 	config, ok := configBlob.Data.(ispec.Image)
 	if !ok {
 		// Should _never_ be reached.
-		return stat, errors.Errorf("[internal error] unknown config blob type: %s", configBlob.Descriptor.MediaType)
+		return stat, fmt.Errorf("[internal error] unknown config blob type: %s", configBlob.Descriptor.MediaType)
 	}
 
 	// TODO: This should probably be moved into separate functions.
@@ -252,8 +272,8 @@ func Stat(ctx context.Context, engine casext.Engine, manifestDescriptor ispec.De
 }
 
 // GenerateBundleManifest creates and writes an mtree of the rootfs in the given
-// bundle path, using the supplied fsEval method
-func GenerateBundleManifest(mtreeName string, bundlePath string, fsEval mtree.FsEval) error {
+// bundle path, using the supplied fsEval method.
+func GenerateBundleManifest(mtreeName string, bundlePath string, fsEval mtree.FsEval) (Err error) {
 	mtreePath := filepath.Join(bundlePath, mtreeName+".mtree")
 	fullRootfsPath := filepath.Join(bundlePath, layer.RootfsName)
 
@@ -265,28 +285,28 @@ func GenerateBundleManifest(mtreeName string, bundlePath string, fsEval mtree.Fs
 	log.Info("computing filesystem manifest ...")
 	dh, err := mtree.Walk(fullRootfsPath, nil, MtreeKeywords, fsEval)
 	if err != nil {
-		return errors.Wrap(err, "generate mtree spec")
+		return fmt.Errorf("generate mtree spec: %w", err)
 	}
 	log.Info("... done")
 
 	flags := os.O_CREATE | os.O_WRONLY | os.O_EXCL
-	fh, err := os.OpenFile(mtreePath, flags, 0644)
+	fh, err := os.OpenFile(mtreePath, flags, 0o644)
 	if err != nil {
-		return errors.Wrap(err, "open mtree")
+		return fmt.Errorf("open mtree: %w", err)
 	}
-	defer fh.Close()
+	defer funchelpers.VerifyClose(&Err, fh)
 
 	log.Debugf("umoci: saving mtree manifest")
 
 	if _, err := dh.WriteTo(fh); err != nil {
-		return errors.Wrap(err, "write mtree")
+		return fmt.Errorf("write mtree: %w", err)
 	}
 
 	return nil
 }
 
 // ParseIdmapOptions sets up the mapping options for Meta, using
-// the arguments specified on the command line
+// the arguments specified on the command line.
 func ParseIdmapOptions(meta *Meta, ctx *cli.Context) error {
 	// We need to set mappings if we're in rootless mode.
 	meta.MapOptions.Rootless = ctx.Bool("rootless")
@@ -294,13 +314,13 @@ func ParseIdmapOptions(meta *Meta, ctx *cli.Context) error {
 		if !ctx.IsSet("uid-map") {
 			if err := ctx.Set("uid-map", fmt.Sprintf("0:%d:1", os.Geteuid())); err != nil {
 				// Should _never_ be reached.
-				return errors.Wrap(err, "[internal error] failure auto-setting rootless --uid-map")
+				return fmt.Errorf("[internal error] failure auto-setting rootless --uid-map: %w", err)
 			}
 		}
 		if !ctx.IsSet("gid-map") {
 			if err := ctx.Set("gid-map", fmt.Sprintf("0:%d:1", os.Getegid())); err != nil {
 				// Should _never_ be reached.
-				return errors.Wrap(err, "[internal error] failure auto-setting rootless --gid-map")
+				return fmt.Errorf("[internal error] failure auto-setting rootless --gid-map: %w", err)
 			}
 		}
 	}
@@ -308,14 +328,14 @@ func ParseIdmapOptions(meta *Meta, ctx *cli.Context) error {
 	for _, uidmap := range ctx.StringSlice("uid-map") {
 		idMap, err := idtools.ParseMapping(uidmap)
 		if err != nil {
-			return errors.Wrapf(err, "failure parsing --uid-map %s", uidmap)
+			return fmt.Errorf("failure parsing --uid-map %s: %w", uidmap, err)
 		}
 		meta.MapOptions.UIDMappings = append(meta.MapOptions.UIDMappings, idMap)
 	}
 	for _, gidmap := range ctx.StringSlice("gid-map") {
 		idMap, err := idtools.ParseMapping(gidmap)
 		if err != nil {
-			return errors.Wrapf(err, "failure parsing --gid-map %s", gidmap)
+			return fmt.Errorf("failure parsing --gid-map %s: %w", gidmap, err)
 		}
 		meta.MapOptions.GIDMappings = append(meta.MapOptions.GIDMappings, idMap)
 	}
